"""LSST the Docs client.
"""

__all__ = ('get_ltd_product_urls', 'get_ltd_product')


async def get_ltd_product_urls(session):
    """Get URLs for LSST the Docs (LTD) products from the LTD Keeper API.

    Parameters
    ----------
    session : `aiohttp.ClientSession`
        Your application's aiohttp client session.
        See http://aiohttp.readthedocs.io/en/stable/client.html.

    Returns
    -------
    product_urls : `list`
        List of product URLs.
    """
    product_url = 'https://keeper.lsst.codes/products/'
    async with session.get(product_url) as response:
        data = await response.json()

    return data['products']


async def get_ltd_product(session, slug=None, url=None):
    """Get the product resource (JSON document) from the LSST the Docs API.

    Parameters
    ----------
    session : `aiohttp.ClientSession`
        Your application's aiohttp client session.
        See http://aiohttp.readthedocs.io/en/stable/client.html.
    slug : `str`, optional
        Slug identfying the product. This is the same as the subdomain.
        For example, ``'ldm-151'`` is the slug for ``ldm-151.lsst.io``.
        A full product URL can be provided instead, see ``url``.
    url : `str`, optional
        The full LTD Keeper URL for the product resource. For example,
        ``'https://keeper.lsst.codes/products/ldm-151'``. The ``slug``
        can be provided instead.

    Returns
    -------
    product : `dict`
        Product dataset. See
        https://ltd-keeper.lsst.io/products.html#get--products-(slug)
        for fields.
    """
    if url is None:
        url = 'https://keeper.lsst.codes/products/{}'.format(slug)

    async with session.get(url) as response:
        data = await response.json()

    return data



"""Tests for the lsstprojectmeta.github.graphql module.
"""

from lsstprojectmeta.github.graphql import GitHubQuery


def test_github_query():
    query = GitHubQuery.load('technote_repo')
    assert str(query) == query.query
    assert query.name == 'technote_repo'
    assert query.query.startswith('query ')



"""Test the lsstprojectmeta.tex.citelink module.
"""

import pytest

from lsstprojectmeta.tex.citelink import (
    CitedsLinker, CitedspLinker, CitepLinker)
from lsstprojectmeta.tex.lsstbib import get_bibliography


@pytest.mark.parametrize(
    'sample,expected',
    # Basic citeds command
    [(r'\citeds{LDM-151}', r'\href{https://ls.st/LDM-151}{LDM-151}'),
     # With an alternative title text
     (r'\citeds[Pipelines Design]{LDM-151}',
      r'\href{https://ls.st/LDM-151}{Pipelines Design}'),
     # Two citeds commands
     (r'\citeds[Pipelines Design]{LDM-151} \citeds{LDM-230}',
      r'\href{https://ls.st/LDM-151}{Pipelines Design} '
      r'\href{https://ls.st/LDM-230}{LDM-230}')])
def test_citeds(sample, expected):
    replace_citeds = CitedsLinker()
    assert replace_citeds(sample) == expected


@pytest.mark.parametrize(
    'sample,expected',
    # Basic citedsp command
    [(r'\citedsp{LDM-151}', r'[\href{https://ls.st/LDM-151}{LDM-151}]'),
     # With an alternative title text
     (r'\citedsp[Pipelines Design]{LDM-151}',
      r'[\href{https://ls.st/LDM-151}{Pipelines Design}]'),
     # Two citedsp commands
     (r'\citedsp[Pipelines Design]{LDM-151} \citedsp{LDM-230}',
      r'[\href{https://ls.st/LDM-151}{Pipelines Design}] '
      r'[\href{https://ls.st/LDM-230}{LDM-230}]')])
def test_citedsp(sample, expected):
    replace_citedsp = CitedspLinker()
    assert replace_citedsp(sample) == expected


@pytest.mark.parametrize(
    'sample,expected',
    # Basic citedp command, single author
    [(r'\citep{1996PASP..108..851S}',
      r'[\href{http://adsabs.harvard.edu/abs/'
      r'1996PASP..108..851S}{Stetson 1996}]'),
     # Basic citep command, multi author
     (r'\citep{2017ApJ...838....5L}',
      r'[\href{http://adsabs.harvard.edu/abs/'
      r'2017ApJ...838....5L}{Leistedt and Hogg 2017}]'),
     # Basic citep command, multi author
     (r'\citep{1996PASP..108..851S, 2017ApJ...838....5L}',
      r'[\href{http://adsabs.harvard.edu/abs/'
      r'1996PASP..108..851S}{Stetson 1996}, '
      r'\href{http://adsabs.harvard.edu/abs/'
      r'2017ApJ...838....5L}{Leistedt and Hogg 2017}]')])
def test_citep(sample, expected):
    db = get_bibliography()
    replace_citep = CitepLinker(bibtex_database=db)
    assert replace_citep(sample) == expected



"""Tests for the tex.commandparser module.
"""

import os
import re

import pytest

from lsstprojectmeta.tex.commandparser import LatexCommand


@pytest.fixture
def ldm_nnn_data():
    data_path = os.path.join(os.path.dirname(__file__), 'data', 'LDM-nnn.tex')
    with open(data_path) as f:
        source = f.read()
    return source


def test_sample_title(ldm_nnn_data):
    """Test parsing the title command from the LDM-nnn name."""
    elements = [
        {'name': 'short_title', 'required': False, 'bracket': '['},
        {'name': 'long_title', 'required': True, 'bracket': '{'}
    ]
    command = LatexCommand('title', *elements)
    parsed_commands = list(command.parse(ldm_nnn_data))

    assert len(parsed_commands) == 1

    parsed = parsed_commands[0]
    assert parsed['long_title'] == 'Title of document'
    assert parsed['short_title'] == 'Short title'

    assert 'short_title' in parsed
    assert 'long_title' in parsed


def test_simple_title():
    """Test parsing the title command for a trivial source sample."""
    sample = (r"Hello world\n\title[Short]{Title}\nFin\n"
              r"\setDocRef{LDM-nnn}\n")
    elements = [
        {'name': 'short_title', 'required': False, 'bracket': '['},
        {'name': 'long_title', 'required': True, 'bracket': '{'}
    ]
    command = LatexCommand('title', *elements)
    parsed = next(command.parse(sample))

    assert parsed['long_title'] == 'Title'
    assert parsed['short_title'] == 'Short'

    assert 'short_title' in parsed
    assert 'long_title' in parsed


def test_no_short_title():
    """Test parsing the title command for a trivial source sample that
    doesn't include a short title.
    """
    sample = ("Hello world\n" + r"\title{Title}" + "\nFin\n"
              r"\setDocRef{LDM-nnn}" + "\n")
    elements = [
        {'name': 'short_title', 'required': False, 'bracket': '['},
        {'name': 'long_title', 'required': True, 'bracket': '{'}
    ]
    command = LatexCommand('title', *elements)
    parsed = next(command.parse(sample))

    assert parsed['long_title'] == 'Title'
    with pytest.raises(KeyError):
        parsed['short_title']

    assert 'short_title' not in parsed
    assert 'long_title' in parsed


def test_no_short_title_v2():
    """Test parsing the title command for a trivial source sample that
    doesn't include a short title, but where the [..]{..} does occur
    elsewhere.
    """
    sample = ("Hello world\n" + r"\title{Title}" + "\nFin\n"
              r"\setDocRef[Trap]{LDM-nnn}" + "\n")
    elements = [
        {'name': 'short_title', 'required': False, 'bracket': '['},
        {'name': 'long_title', 'required': True, 'bracket': '{'}
    ]
    command = LatexCommand('title', *elements)
    parsed = next(command.parse(sample))

    assert parsed['long_title'] == 'Title'
    with pytest.raises(KeyError):
        parsed['short_title']

    assert 'short_title' not in parsed
    assert 'long_title' in parsed


@pytest.mark.parametrize(
    'command,sample',
    [('title', '% hello\n' + r'\title{Hello world}'),
     ('title', '% hello\n' + r'\title{Hello world}' + '\n'
         r'\titlename{Imposter}')])
def test_command_regex(command, sample):
    """Test the regex make my LatexCommand._make_command_regex to ensure that
    it detects the command and not look-alike latex commands. Each sample
    should have only one detection.
    """
    command_regex = LatexCommand._make_command_regex(command)
    matches = re.findall(command_regex, sample)
    assert len(matches) == 1


def test_optional_bracket():
    """Test parsing a command where the sole token has no brackets around it.
    """
    sample = (
        "Hello.\n\n"
        r"\input test.tex"
        "\n\nMore content."
    )

    command = LatexCommand(
        'input',
        {'name': 'filename', 'required': True, 'bracket': '{'}
    )
    parsed = next(command.parse(sample))
    assert parsed['filename'] == 'test.tex'



"""Reduce technote projects into JSON-LD metadata.
"""

__all__ = ('process_sphinx_technote', 'reduce_technote_metadata',
           'download_metadata_yaml', 'NotSphinxTechnoteError')

import datetime
import logging

import aiohttp
import yaml

from ..github.urls import (parse_repo_slug_from_url, make_raw_content_url,
                           normalize_repo_root_url)
from ..github.graphql import github_request, GitHubQuery


async def process_sphinx_technote(session, github_api_token, ltd_product_data,
                                  mongo_collection=None):
    """Extract, transform, and load Sphinx-based technote metadata.

    Parameters
    ----------
    session : `aiohttp.ClientSession`
        Your application's aiohttp client session.
        See http://aiohttp.readthedocs.io/en/stable/client.html.
    github_api_token : `str`
        A GitHub personal API token. See the `GitHub personal access token
        guide`_.
    ltd_product_data : `dict`
        Contents of ``metadata.yaml``, obtained via `download_metadata_yaml`.
        Data for this technote from the LTD Keeper API
        (``GET /products/<slug>``). Usually obtained via
        `lsstprojectmeta.ltd.get_ltd_product`.
    mongo_collection : `motor.motor_asyncio.AsyncIOMotorCollection`, optional
        MongoDB collection. This should be the common MongoDB collection for
        LSST projectmeta JSON-LD records. If provided, ths JSON-LD is upserted
        into the MongoDB collection.

    Returns
    -------
    metadata : `dict`
        JSON-LD-formatted dictionary.

    Raises
    ------
    NotSphinxTechnoteError
        Raised when the LTD product cannot be interpreted as a Sphinx-based
        technote project because it's missing a metadata.yaml file in its
        GitHub repository. This implies that the LTD product *could* be of a
        different format.

    .. `GitHub personal access token guide`: https://ls.st/41d
    """
    logger = logging.getLogger(__name__)

    github_url = ltd_product_data['doc_repo']
    github_url = normalize_repo_root_url(github_url)
    repo_slug = parse_repo_slug_from_url(github_url)

    try:
        metadata_yaml = await download_metadata_yaml(session, github_url)
    except aiohttp.ClientResponseError as err:
        # metadata.yaml not found; probably not a Sphinx technote
        logger.debug('Tried to download %s\'s metadata.yaml, got status %d',
                     ltd_product_data['slug'], err.code)
        raise NotSphinxTechnoteError()

    # Extract data from the GitHub API
    github_query = GitHubQuery.load('technote_repo')
    github_variables = {
        "orgName": repo_slug.owner,
        "repoName": repo_slug.repo
    }
    github_data = await github_request(session, github_api_token,
                                       query=github_query,
                                       variables=github_variables)

    try:
        jsonld = reduce_technote_metadata(
            github_url, metadata_yaml, github_data, ltd_product_data)
    except Exception as exception:
        message = "Issue building JSON-LD for technote %s"
        logger.exception(message, github_url, exception)
        raise

    if mongo_collection is not None:
        await _upload_to_mongodb(mongo_collection, jsonld)

    logger.info('Ingested technote %s into MongoDB', github_url)

    return jsonld


def reduce_technote_metadata(github_url, metadata, github_data,
                             ltd_product_data):
    """Reduce a technote project's metadata from multiple sources into a
    single JSON-LD resource.

    Parameters
    ----------
    github_url : `str`
        URL of the technote's GitHub repository.
    metadata : `dict`
        The parsed contents of ``metadata.yaml`` found in a technote's
        repository.
    github_data : `dict`
        The contents of the ``technote_repo`` GitHub GraphQL API query.
    ltd_product_data : `dict`
        JSON dataset for the technote corresponding to the
        ``/products/<product>`` of LTD Keeper.

    Returns
    -------
    metadata : `dict`
        JSON-LD-formatted dictionary.

    .. `GitHub personal access token guide`: https://ls.st/41d
    """
    repo_slug = parse_repo_slug_from_url(github_url)

    # Initialize a schema.org/Report and schema.org/SoftwareSourceCode
    # linked data resource
    jsonld = {
        '@context': [
            "https://raw.githubusercontent.com/codemeta/codemeta/2.0-rc/"
            "codemeta.jsonld",
            "http://schema.org"],
        '@type': ['Report', 'SoftwareSourceCode'],
        'codeRepository': github_url
    }

    if 'url' in metadata:
        url = metadata['url']
    elif 'published_url' in ltd_product_data:
        url = ltd_product_data['published_url']
    else:
        raise RuntimeError('No identifying url could be found: '
                           '{}'.format(github_url))
    jsonld['@id'] = url
    jsonld['url'] = url

    if 'series' in metadata and 'serial_number' in metadata:
        jsonld['reportNumber'] = '{series}-{serial_number}'.format(**metadata)
    else:
        raise RuntimeError('No reportNumber: {}'.format(github_url))

    if 'doc_title' in metadata:
        jsonld['name'] = metadata['doc_title']

    if 'description' in metadata:
        jsonld['description'] = metadata['description']

    if 'authors' in metadata:
        jsonld['author'] = [{'@type': 'Person', 'name': author_name}
                            for author_name in metadata['authors']]

    if 'last_revised' in metadata:
        # Prefer getting the 'last_revised' date from metadata.yaml
        # since it's considered an override.
        jsonld['dateModified'] = datetime.datetime.strptime(
            metadata['last_revised'],
            '%Y-%m-%d')
    else:
        # Fallback to parsing the date of the last commit to the
        # default branch on GitHub (usually `master`).
        try:
            _repo_data = github_data['data']['repository']
            _master_data = _repo_data['defaultBranchRef']
            jsonld['dateModified'] = datetime.datetime.strptime(
                _master_data['target']['committedDate'],
                '%Y-%m-%dT%H:%M:%SZ')
        except KeyError:
            pass

    try:
        _license_data = github_data['data']['repository']['licenseInfo']
        _spdxId = _license_data['spdxId']
        if _spdxId is not None:
            _spdx_url = 'https://spdx.org/licenses/{}.html'.format(_spdxId)
            jsonld['license'] = _spdx_url
    except KeyError:
        pass

    try:
        # Find the README(|.md|.rst|*) file in the repo root
        _master_data = github_data['data']['repository']['defaultBranchRef']
        _files = _master_data['target']['tree']['entries']
        for _node in _files:
            filename = _node['name']
            normalized_filename = filename.lower()
            if normalized_filename.startswith('readme'):
                readme_url = make_raw_content_url(repo_slug, 'master',
                                                  filename)
                jsonld['readme'] = readme_url
                break
    except KeyError:
        pass

    # Assume Travis is the CI service (always true at the moment)
    travis_url = 'https://travis-ci.org/{}'.format(repo_slug.full)
    jsonld['contIntegration'] = travis_url

    return jsonld


async def download_metadata_yaml(session, github_url):
    """Download the metadata.yaml file from a technote's GitHub repository.
    """
    metadata_yaml_url = _build_metadata_yaml_url(github_url)
    async with session.get(metadata_yaml_url) as response:
        response.raise_for_status()
        yaml_data = await response.text()
    return yaml.safe_load(yaml_data)


def _build_metadata_yaml_url(github_url):
    """Compute the URL to the raw metadata.yaml resource given the technote's
    GitHub repository URL.

    Parameters
    ----------
    github_url : `str`
        URL of the technote's GitHub repository.

    Returns
    -------
    metadata_yaml_url : `str`
        metadata.yaml URL (using the ``raw.githubusercontent.com`` domain).
    """
    repo_slug = parse_repo_slug_from_url(github_url)
    return make_raw_content_url(repo_slug, 'master', 'metadata.yaml')


async def _upload_to_mongodb(collection, jsonld):
    """Upsert the technote resource into the projectmeta MongoDB collection.

    Parameters
    ----------
    collection : `motor.motor_asyncio.AsyncIOMotorCollection`
        The MongoDB collection.
    jsonld : `dict`
        The JSON-LD document that reprsents the technote resource.
    """
    document = {
        'data': jsonld
    }
    query = {
        'data.reportNumber': jsonld['reportNumber']
    }
    await collection.update(query, document, upsert=True, multi=False)


class NotSphinxTechnoteError(Exception):
    """Exception indicating that an LSST the Docs product cannot be a
    Sphinx-formatted technote and that a different format should be tried.
    """



"""Tests for the lsstprojectmeta.git.timestamp.
"""

from datetime import datetime
import os

import pytest

from lsstprojectmeta.git.timestamp import (
    read_git_commit_timestamp_for_file,
    _iter_filepaths_with_extension,
    get_content_commit_date)


def test_git_commit_timestamp_for_file():
    """Smoke-test read_git_commit_timestamp_for_file with README.rst in
    lsstprojectmeta's own Git repo.
    """
    test_dir = os.path.dirname(__file__)
    readme_path = os.path.abspath(os.path.join(test_dir, '..', 'README.rst'))
    timestamp = read_git_commit_timestamp_for_file(readme_path)
    assert isinstance(timestamp, datetime)


def test_git_commit_timestamp_for_file_nonexistent():
    """Smoke-test read_git_commit_timestamp_for_file with README.rst in
    lsstprojectmeta's own Git repo.
    """
    path = 'doesnt_exist.txt'
    with pytest.raises(IOError):
        read_git_commit_timestamp_for_file(path)


def test_iter_filepaths_with_extension():
    repo_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
    filepaths = _iter_filepaths_with_extension('py', root_dir=repo_dir)
    assert os.path.join('tests', 'test_git_timestamp.py') in filepaths
    assert 'README.rst' not in filepaths


def test_get_project_content_commit_date():
    """Smoke test using the lsstprojectmeta repo.
    """
    repo_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
    commit_date = get_content_commit_date(('rst', 'py'), root_dir=repo_dir)
    assert isinstance(commit_date, datetime)

    # Verify file search is case sensitive
    with pytest.raises(RuntimeError):
        get_content_commit_date(('RST', 'PY'), root_dir=repo_dir)



"""Tests for the lsstprojectmeta.tex.normalizer module.
"""

import os
import re

import pytest

import lsstprojectmeta.tex.normalizer as normalizer


def test_remove_comments_abstract():
    sample = (r"\setDocAbstract{%" + '\n'
              " The LSST Data Management System (DMS) is a set of services\n"
              " employing a variety of software components running on\n"
              " computational and networking infrastructure that combine to\n"
              " deliver science data products to the observatory's users and\n"
              " support observatory operations.  This document describes the\n"
              " components, their service instances, and their deployment\n"
              " environments as well as the interfaces among them, the rest\n"
              " of the LSST system, and the outside world.\n"
              "}")
    expected = (
        r"\setDocAbstract{" + '\n'
        " The LSST Data Management System (DMS) is a set of services\n"
        " employing a variety of software components running on\n"
        " computational and networking infrastructure that combine to\n"
        " deliver science data products to the observatory's users and\n"
        " support observatory operations.  This document describes the\n"
        " components, their service instances, and their deployment\n"
        " environments as well as the interfaces among them, the rest\n"
        " of the LSST system, and the outside world.\n"
        "}")
    assert normalizer.remove_comments(sample) == expected


def test_escaped_remove_comments():
    """Test remove_comments where a "%" is escaped."""
    sample = r"The uncertainty is 5\%.  % a comment"
    expected = r"The uncertainty is 5\%.  "
    assert normalizer.remove_comments(sample) == expected


def test_single_line_remove_comments():
    sample = "This is content.  % a comment"
    expected = "This is content.  "
    assert normalizer.remove_comments(sample) == expected


def test_remove_single_line_trailing_whitespace():
    sample = "This is content.    "
    expected = "This is content."
    assert normalizer.remove_trailing_whitespace(sample) == expected


def test_multi_line_trailing_whitespace():
    sample = ("First line.    \n"
              "Second line. ")
    expected = ("First line.\n"
                "Second line.")
    assert normalizer.remove_trailing_whitespace(sample) == expected


def test_read_tex_file():
    project_dir = os.path.join(os.path.dirname(__file__), 'data', 'texinputs')
    root_filepath = os.path.join(project_dir, 'LDM-nnn.tex')
    tex_source = normalizer.read_tex_file(root_filepath)

    # verify that input'd and include'd content is present
    assert re.search(r'\\setDocAbstract', tex_source) is not None
    assert re.search(r'\\section{Introduction}', tex_source) is not None


def test_replace_macros():
    sample = (
        r"\def \product {Data Management}" + "\n"
        r"\title    [Test Plan]  { \product\ Test Plan}" + "\n"
        r"\setDocAbstract {" + "\n"
        r"This is the  Test Plan for \product.}")

    expected = (
        r"\def Data Management {Data Management}" + "\n"
        r"\title    [Test Plan]  { Data Management Test Plan}" + "\n"
        r"\setDocAbstract {" + "\n"
        r"This is the  Test Plan for Data Management.}")

    macros = {r'\product': 'Data Management'}
    tex_source = normalizer.replace_macros(sample, macros)
    assert re.search(r'\\product', sample) is not None  # sanity check
    assert re.search(r'\\product', tex_source) is None
    assert tex_source == expected


@pytest.mark.parametrize(
    'sample,expected',
    [(r'\input{file.tex}', 'file.tex'),
     (r'\input{dirname/file.tex}', 'dirname/file.tex'),
     (r'\input {file}%', 'file'),
     (r'\input file%', 'file'),
     (r'\input file' + '\n', 'file'),
     (r'\input file ' + '\n', 'file'),
     (r'\include{file.tex}', 'file.tex'),
     (r'\include{dirname/file.tex}', 'dirname/file.tex'),
     (r'\include {file}%', 'file'),
     (r'\include file%', 'file'),
     (r'\include file' + '\n', 'file'),
     (r'\include file' + ' \n', 'file')])
def test_input_include_pattern(sample, expected):
    match = re.search(normalizer.input_include_pattern, sample)
    assert match.group('filename') == expected


def test_non_inputs():
    r"""Test for patterns like ``\inputData{XYZ}`` that have in the past been
    detected as an ``\input`` command.
    """
    sample = r'\newcommand{\inputData}[1]{\texttt{#1}}'
    match = re.search(normalizer.input_include_pattern, sample)
    assert match is None



"""Pandoc conversion helper functions.
"""

__all__ = ['convert_text', 'ensure_pandoc']

import functools
import logging

import pypandoc

from ..tex.lsstmacros import LSSTDOC_MACROS


def ensure_pandoc(func):
    """Decorate a function that uses pypandoc to ensure that pandoc is
    installed if necessary.
    """
    logger = logging.getLogger(__name__)

    @functools.wraps(func)
    def _install_and_run(*args, **kwargs):
        try:
            # First try to run pypandoc function
            result = func(*args, **kwargs)
        except OSError:
            # Install pandoc and retry
            message = "pandoc needed but not found. Now installing it for you."
            logger.warning(message)
            # This version of pandoc is known to be compatible with both
            # pypandoc.download_pandoc and the functionality that
            # lsstprojectmeta needs. Travis CI tests are useful for ensuring
            # download_pandoc works.
            pypandoc.download_pandoc(version='1.19.1')
            logger.debug("pandoc download complete")

            result = func(*args, **kwargs)

        return result

    return _install_and_run


@ensure_pandoc
def convert_text(content, from_fmt, to_fmt, deparagraph=False, mathjax=False,
                 smart=True, extra_args=None):
    """Convert text from one markup format to another using pandoc.

    This function is a thin wrapper around `pypandoc.convert_text`.

    Parameters
    ----------
    content : `str`
        Original content.

    from_fmt : `str`
        Format of the original ``content``. Format identifier must be one of
        those known by Pandoc. See https://pandoc.org/MANUAL.html for details.

    to_fmt : `str`
        Output format for the content.

    deparagraph : `bool`, optional
        If `True`, then the
        `lsstprojectmeta.pandoc.filters.deparagraph.deparagraph` filter is
        used to remove paragraph (``<p>``, for example) tags around a single
        paragraph of content. That filter does not affect content that
        consists of multiple blocks (several paragraphs, or lists, for
        example). Default is `False`.

        For example, **without** this filter Pandoc will convert
        the string ``"Title text"`` to ``"<p>Title text</p>"`` in HTML. The
        paragraph tags aren't useful if you intend to wrap the converted
        content in different tags, like ``<h1>``, using your own templating
        system.

        **With** this filter, Pandoc will convert the string ``"Title text"``
        to ``"Title text"`` in HTML.

    mathjax : `bool`, optional
        If `True` then Pandoc will markup output content to work with MathJax.
        Default is False.

    smart : `bool`, optional
        If `True` (default) then ascii characters will be converted to unicode
        characters like smart quotes and em dashes.

    extra_args : `list`, optional
        Sequence of Pandoc arguments command line arguments (such as
        ``'--normalize'``). The ``deparagraph``, ``mathjax``, and ``smart``
        arguments are convenience arguments that are equivalent to items
        in ``extra_args``.

    Returns
    -------
    output : `str`
        Content in the output (``to_fmt``) format.

    Notes
    -----
    This function will automatically install Pandoc if it is not available.
    See `ensure_pandoc`.
    """
    logger = logging.getLogger(__name__)

    if extra_args is not None:
        extra_args = list(extra_args)
    else:
        extra_args = []

    if mathjax:
        extra_args.append('--mathjax')

    if smart:
        extra_args.append('--smart')

    if deparagraph:
        extra_args.append('--filter=lsstprojectmeta-deparagraph')

    extra_args.append('--wrap=none')

    # de-dupe extra args
    extra_args = set(extra_args)

    logger.debug('Running pandoc from %s to %s with extra_args %s',
                 from_fmt, to_fmt, extra_args)

    output = pypandoc.convert_text(content, to_fmt, format=from_fmt,
                                   extra_args=extra_args)
    return output


def convert_lsstdoc_tex(
        content, to_fmt, deparagraph=False, mathjax=False,
        smart=True, extra_args=None):
    """Convert lsstdoc-class LaTeX to another markup format.

    This function is a thin wrapper around `convert_text` that automatically
    includes common lsstdoc LaTeX macros.

    Parameters
    ----------
    content : `str`
        Original content.

    to_fmt : `str`
        Output format for the content (see https://pandoc.org/MANUAL.html).
        For example, 'html5'.

    deparagraph : `bool`, optional
        If `True`, then the
        `lsstprojectmeta.pandoc.filters.deparagraph.deparagraph` filter is
        used to remove paragraph (``<p>``, for example) tags around a single
        paragraph of content. That filter does not affect content that
        consists of multiple blocks (several paragraphs, or lists, for
        example). Default is `False`.

        For example, **without** this filter Pandoc will convert
        the string ``"Title text"`` to ``"<p>Title text</p>"`` in HTML. The
        paragraph tags aren't useful if you intend to wrap the converted
        content in different tags, like ``<h1>``, using your own templating
        system.

        **With** this filter, Pandoc will convert the string ``"Title text"``
        to ``"Title text"`` in HTML.

    mathjax : `bool`, optional
        If `True` then Pandoc will markup output content to work with MathJax.
        Default is False.

    smart : `bool`, optional
        If `True` (default) then ascii characters will be converted to unicode
        characters like smart quotes and em dashes.

    extra_args : `list`, optional
        Sequence of Pandoc arguments command line arguments (such as
        ``'--normalize'``). The ``deparagraph``, ``mathjax``, and ``smart``
        arguments are convenience arguments that are equivalent to items
        in ``extra_args``.

    Returns
    -------
    output : `str`
        Content in the output (``to_fmt``) format.

    Notes
    -----
    This function will automatically install Pandoc if it is not available.
    See `ensure_pandoc`.
    """
    augmented_content = '\n'.join((LSSTDOC_MACROS, content))
    return convert_text(
        augmented_content, 'latex', to_fmt,
        deparagraph=deparagraph, mathjax=mathjax,
        smart=smart, extra_args=extra_args)



"""TeX source scraping.
"""

__all__ = ['get_macros', 'get_def_macros', 'get_newcommand_macros']

import re

from .commandparser import LatexCommand

# Regular expressions for "\def \name {content}"
# Expects the entire command to be on one line.
DEF_PATTERN = re.compile(
    r"\\def\s*"  # def command with optional whitespace
    r"(?P<name>\\[a-zA-Z]*?)\s*"  # macro name with optional whitespace
    r"{(?P<content>.*?)}")  # macro contents


def get_macros(tex_source):
    r"""Get all macro definitions from TeX source, supporting multiple
    declaration patterns.

    Parameters
    ----------
    tex_source : `str`
        TeX source content.

    Returns
    -------
    macros : `dict`
        Keys are macro names (including leading ``\``) and values are the
        content (as `str`) of the macros.

    Notes
    -----
    This function uses the following function to scrape macros of different
    types:

    - `get_def_macros`
    - `get_newcommand_macros`

    This macro scraping has the following caveats:

    - Macro definition (including content) must all occur on one line.
    - Macros with arguments are not supported.
    """
    macros = {}
    macros.update(get_def_macros(tex_source))
    macros.update(get_newcommand_macros(tex_source))
    return macros


def get_def_macros(tex_source):
    r"""Get all ``\def`` macro definition from TeX source.

    Parameters
    ----------
    tex_source : `str`
        TeX source content.

    Returns
    -------
    macros : `dict`
        Keys are macro names (including leading ``\``) and values are the
        content (as `str`) of the macros.

    Notes
    -----
    ``\def`` macros with arguments are not supported.
    """
    macros = {}
    for match in DEF_PATTERN.finditer(tex_source):
        macros[match.group('name')] = match.group('content')
    return macros


def get_newcommand_macros(tex_source):
    r"""Get all ``\newcommand`` macro definition from TeX source.

    Parameters
    ----------
    tex_source : `str`
        TeX source content.

    Returns
    -------
    macros : `dict`
        Keys are macro names (including leading ``\``) and values are the
        content (as `str`) of the macros.

    Notes
    -----
    ``\newcommand`` macros with arguments are not supported.
    """
    macros = {}
    command = LatexCommand(
        'newcommand',
        {'name': 'name', 'required': True, 'bracket': '{'},
        {'name': 'content', 'required': True, 'bracket': '{'})

    for macro in command.parse(tex_source):
        macros[macro['name']] = macro['content']

    return macros



"""Ad hoc tests of the LsstLatexDoc class. Other test modules rigorously verify
LsstLatexDoc against sample documents.
"""

from pybtex.database import BibliographyData
import pytest

from lsstprojectmeta.tex.lsstdoc import LsstLatexDoc


def test_no_short_title():
    """title without a short title."""
    sample = r"\title{Title}"
    lsstdoc = LsstLatexDoc(sample)
    assert lsstdoc.title == "Title"


def test_title_variations():
    """Test variations on the title command's formatting."""
    # Test with whitespace in title command
    input_txt = r"\title    [Test Plan]  { \product ~Test Plan}"
    lsstdoc = LsstLatexDoc(input_txt)
    assert lsstdoc.title == r"\product ~Test Plan"
    assert lsstdoc.short_title == "Test Plan"


def test_author_variations():
    """Test variations on the author command's formatting."""
    input_txt = (r"\author   {William O'Mullane, Mario Juric, "
                 r"Frossie Economou}"
                 r"                  % the author(s)")
    lsstdoc = LsstLatexDoc(input_txt)
    assert lsstdoc.authors == ["William O'Mullane",
                               "Mario Juric",
                               "Frossie Economou"]


def test_author_list_amanda():
    """Test author list parsing where one author's name is Amanda.
    """
    input_txt = (
        r"\author   {William O'Mullane, John Swinbank, Leanne Guy, "
        r"Amanda Bauer}"
    )
    expected = [
        "William O'Mullane",
        "John Swinbank",
        "Leanne Guy",
        "Amanda Bauer"
    ]
    lsstdoc = LsstLatexDoc(input_txt)
    assert lsstdoc.authors == expected


def test_author_list_and():
    input_txt = r"\author{A.~Author, B.~Author, and C.~Author}"
    expected = ['A. Author', 'B. Author', 'C. Author']
    lsstdoc = LsstLatexDoc(input_txt)
    assert lsstdoc.authors == expected


def test_handle_variations():
    """Test variations on the handle command's formatting."""
    input_txt = r"\setDocRef      {LDM-503} % the reference code "
    lsstdoc = LsstLatexDoc(input_txt)
    assert lsstdoc.handle == "LDM-503"


def test_abstract_variations():
    """Test variations on the abstract command's formatting."""
    input_txt = (r"\setDocAbstract {" + "\n"
                 r"This is the  Test Plan for \product. In it we define terms "
                 r"associated with testing and further test specifications "
                 r"for specific items.}")
    expected_abstract = (
        r"This is the  Test Plan for \product. In it we define terms "
        r"associated with testing and further test specifications for "
        r"specific items."
    )
    lsstdoc = LsstLatexDoc(input_txt)
    assert lsstdoc.abstract == expected_abstract


@pytest.mark.parametrize(
    'sample, expected',
    [(r'\documentclass[DM,lsstdraft,toc]{lsstdoc}', True),
     (r'\documentclass[DM,toc]{lsstdoc}', False),
     (r'\documentclass[DM, lsstdraft, toc]{lsstdoc}', True)])
def test_is_draft(sample, expected):
    lsstdoc = LsstLatexDoc(sample)
    assert lsstdoc.is_draft == expected


def test_html_title():
    sample = r"\title{``Complex'' title \textit{like} $1+2$}"
    expected = ('“Complex” title <em>like</em> '
                '<span class="math inline">1\u2005+\u20052</span>\n')
    lsstdoc = LsstLatexDoc(sample)
    converted = lsstdoc.html_title
    assert converted == expected


def test_default_load_bib_db():
    """Test that the common lsst-texmf bibliographies are always loaded.
    """
    lsstdoc = LsstLatexDoc('')
    assert isinstance(lsstdoc.bib_db, BibliographyData)



"""Flexible LaTeX command parsing.
"""

__all__ = ['LatexCommand', 'ParsedCommand']

import logging
import re


class LatexCommand(object):
    r"""Definition of a LaTeX command's syntax that is used for parsing that
    command's content from a LaTeX source document.

    Parameters
    ----------
    name : `str`
        Name of the LaTeX command. For example, the name of the ``'\title'``
        command is ``'title'`` (without the backslash prefix).
    *elements : `dict`
        Each position element is a dictionary describing that element of the
        command. These elements are provided in the order they appear in
        the command syntax.

        The keys of each dictionary are:

        - ``bracket``: The element's bracket style. Can be ``'['`` or ``'{'``.
        - ``required`` (optional field): `False` if the field is optional.
          `True` if required. Default is `True`.
        - ``name`` (optional field): Name of the field.
    """

    _brackets = {'[': ']', '{': '}'}

    def __init__(self, name, *elements):
        super().__init__()
        self._logger = logging.getLogger(__name__)
        self.name = name
        # Should add validation to elements
        self.elements = []
        for i, element in enumerate(elements):
            base_element = dict(bracket='{', required=True, name=None)
            base_element.update(element)
            base_element['index'] = i
            self.elements.append(base_element)

    def parse(self, source):
        """Parse command content from the LaTeX source.

        Parameters
        ----------
        source : `str`
            The full source of the tex document.

        Yields
        ------
        parsed_command : `ParsedCommand`
            Yields parsed commands instances for each occurence of the command
            in the source.
        """
        command_regex = self._make_command_regex(self.name)
        for match in re.finditer(command_regex, source):
            self._logger.debug(match)
            start_index = match.start(0)
            yield self._parse_command(source, start_index)

    @staticmethod
    def _make_command_regex(name):
        r"""Given a command name, build a regular expression to detect that
        command in TeX source.

        The regular expression is designed to discern "\title{..}" from
        "\titlename{..}". It does this by ensuring that the command is
        followed by a whitespace character, argument brackets, or a comment
        character.

        Parameters
        ----------
        name : `str`
            Name of the command (with a backslash prefix).

        Returns
        -------
        regex : `str`
            Regular expression pattern for detecting the command.
        """
        return r'\\' + name + r'(?:[\s{[%])'

    def _parse_command(self, source, start_index):
        """Parse a single command.

        Parameters
        ----------
        source : `str`
            The full source of the tex document.
        start_index : `int`
            Character index in ``source`` where the command begins.

        Returns
        -------
        parsed_command : `ParsedCommand`
            The parsed command from the source at the given index.
        """
        parsed_elements = []

        # Index of the parser in the source
        running_index = start_index

        for element in self.elements:
            opening_bracket = element['bracket']
            closing_bracket = self._brackets[opening_bracket]

            # Find the opening bracket.
            element_start = None
            element_end = None
            for i, c in enumerate(source[running_index:], start=running_index):
                if c == element['bracket']:
                    element_start = i
                    break
                elif c == '\n':
                    # No starting bracket on the line.
                    if element['required'] is True:
                        # Try to parse a single single-word token after the
                        # command, like '\input file'
                        content = self._parse_whitespace_argument(
                            source[running_index:],
                            self.name)
                        return ParsedCommand(
                            self.name,
                            [{'index': element['index'],
                              'name': element['name'],
                              'content': content.strip()}],
                            start_index,
                            source[start_index:i])
                    else:
                        # Give up on finding an optional element
                        break

            # Handle cases when the opening bracket is never found.
            if element_start is None and element['required'] is False:
                # Optional element not found. Continue to next element,
                # not advancing the running_index of the parser.
                continue
            elif element_start is None and element['required'] is True:
                message = ('Parsing command {0} at index {1:d}, '
                           'did not detect element {2:d}'.format(
                               self.name,
                               start_index,
                               element['index']))
                raise CommandParserError(message)

            # Find the closing bracket, keeping track of the number of times
            # the same type of bracket was opened and closed.
            balance = 1
            for i, c in enumerate(source[element_start + 1:],
                                  start=element_start + 1):
                if c == opening_bracket:
                    balance += 1
                elif c == closing_bracket:
                    balance -= 1

                if balance == 0:
                    element_end = i
                    break

            if balance > 0:
                message = ('Parsing command {0} at index {1:d}, '
                           'did not find closing bracket for required '
                           'command element {2:d}'.format(
                               self.name,
                               start_index,
                               element['index']))
                raise CommandParserError(message)

            # Package the parsed element's content.
            element_content = source[element_start + 1:element_end]
            parsed_element = {
                'index': element['index'],
                'name': element['name'],
                'content': element_content.strip()
            }
            parsed_elements.append(parsed_element)

            running_index = element_end + 1

        command_source = source[start_index:running_index]
        parsed_command = ParsedCommand(self.name, parsed_elements,
                                       start_index, command_source)
        return parsed_command

    @staticmethod
    def _parse_whitespace_argument(source, name):
        r"""Attempt to parse a single token on the first line of this source.

        This method is used for parsing whitespace-delimited arguments, like
        ``\input file``. The source should ideally contain `` file`` along
        with a newline character.

        >>> source = 'Line 1\n' r'\input test.tex' '\nLine 2'
        >>> LatexCommand._parse_whitespace_argument(source, 'input')
        'test.tex'

        Bracket delimited arguments (``\input{test.tex}``) are handled in
        the normal logic of `_parse_command`.
        """
        # First match the command name itself so that we find the argument
        # *after* the command
        command_pattern = r'\\(' + name + r')(?:[\s{[%])'
        command_match = re.search(command_pattern, source)
        if command_match is not None:
            # Trim `source` so we only look after the command
            source = source[command_match.end(1):]

        # Find the whitespace-delimited argument itself.
        pattern = r'(?P<content>\S+)(?:[ %\t\n]+)'
        match = re.search(pattern, source)
        if match is None:
            message = (
                'When parsing {}, did not find whitespace-delimited command '
                'argument'
            )
            raise CommandParserError(message.format(name))
        content = match.group('content')
        content.strip()
        return content


class ParsedCommand(object):
    r"""Contents of a parsed LaTeX command.

    Parameters
    ----------
    name : `str`
        Name of the LaTeX command. For example, the name of the ``'\title'``
        command is ``'title'`` (without the backslash prefix).
    parsed_elements : `list`
        Parsed command elements. Each item is a `dict` with keys:

        - ``index``: `int` index of the element in the `LatexCommand` command
          definition.
        - ``name``: `str` name of the element in the `LatexCommand` command
          definition, or `None` if the element is not named.
        - ``content``: `str` content of the element (tex source).
    start_index : `int`
        Character index in the ``full_source`` where the command begins.
    command_source : `str`
        The full source of the parsed LaTeX command. This can be used to
        replace the full command in the source document.
    """

    def __init__(self, name, parsed_elements, start_index, command_source):
        super().__init__()
        self.name = name
        self.start_index = start_index
        self._parsed_elements = parsed_elements
        self.command_source = command_source

    def __getitem__(self, key):
        element = self._get_element(key)
        return element['content']

    def __contains__(self, key):
        try:
            self._get_element(key)
        except KeyError:
            return False

        return True

    def _get_element(self, key):
        element = None

        if isinstance(key, int):
            # Get by element index
            for _element in self._parsed_elements:
                if _element['index'] == key:
                    element = _element
                    break
        else:
            # Get by element name
            for _element in self._parsed_elements:
                if _element['name'] == key:
                    element = _element
                    break

        if element is None:
            message = 'Key {} not found'.format(key)
            raise KeyError(message)
        return element


class CommandParserError(Exception):
    """Error raised when parsing commands from LaTeX source with the
    `CommandParser` class.
    """



"""APIs for getting and working with the BibTeX databases in lsst-texmf.
"""

__all__ = ['get_lsst_bibtex', 'get_bibliography', 'get_url_from_entry',
           'NoEntryUrlError', 'get_authoryear_from_entry', 'AuthorYearError']

import asyncio
import logging
import os

from aiohttp import ClientSession
import pybtex.database

# https://lsst-texmf.lsst.io/lsstdoc.html#bibliographies
KNOWN_LSSTTEXMF_BIB_NAMES = ('lsst', 'lsst-dm', 'refs', 'books', 'refs_ads')


# Cache of bibtex file content, keyed by name (see KNOWN_LSSTTEXMF_BIB_NAMES).
_LSSTTEXMF_BIB_CACHE = {}


async def _download_text(url, session):
    """Asynchronously request a URL and get the encoded text content of the
    body.

    Parameters
    ----------
    url : `str`
        URL to download.
    session : `aiohttp.ClientSession`
        An open aiohttp session.

    Returns
    -------
    content : `str`
        Content downloaded from the URL.
    """
    logger = logging.getLogger(__name__)
    async with session.get(url) as response:
        # aiohttp decodes the content to a Python string
        logger.info('Downloading %r', url)
        return await response.text()


async def _download_lsst_bibtex(bibtex_names):
    """Asynchronously download a set of lsst-texmf BibTeX bibliographies from
    GitHub.

    Parameters
    ----------
    bibtex_names : sequence of `str`
        Names of lsst-texmf BibTeX files to download. For example:

        .. code-block:: python

           ['lsst', 'lsst-dm', 'refs', 'books', 'refs_ads']

    Returns
    -------
    bibtexs : `list` of `str`
        List of BibTeX file content, in the same order as ``bibtex_names``.
    """
    blob_url_template = (
        'https://raw.githubusercontent.com/lsst/lsst-texmf/master/texmf/'
        'bibtex/bib/{name}.bib'
    )
    urls = [blob_url_template.format(name=name) for name in bibtex_names]

    tasks = []
    async with ClientSession() as session:
        for url in urls:
            task = asyncio.ensure_future(_download_text(url, session))
            tasks.append(task)

        return await asyncio.gather(*tasks)


def get_lsst_bibtex(bibtex_filenames=None):
    """Get content of lsst-texmf bibliographies.

    BibTeX content is downloaded from GitHub (``master`` branch of
    https://github.com/lsst/lsst-texmf or retrieved from an in-memory cache.

    Parameters
    ----------
    bibtex_filenames : sequence of `str`, optional
        List of lsst-texmf BibTeX files to retrieve. These can be the filenames
        of lsst-bibtex files (for example, ``['lsst.bib', 'lsst-dm.bib']``)
        or names without an extension (``['lsst', 'lsst-dm']``). The default
        (recommended) is to get *all* lsst-texmf bibliographies:

        .. code-block:: python

           ['lsst', 'lsst-dm', 'refs', 'books', 'refs_ads']

    Returns
    -------
    bibtex : `dict`
        Dictionary with keys that are bibtex file names (such as ``'lsst'``,
        ``'lsst-dm'``). Values are the corresponding bibtex file content
        (`str`).
    """
    logger = logging.getLogger(__name__)

    if bibtex_filenames is None:
        # Default lsst-texmf bibliography files
        bibtex_names = KNOWN_LSSTTEXMF_BIB_NAMES
    else:
        # Sanitize filenames (remove extensions, path)
        bibtex_names = []
        for filename in bibtex_filenames:
            name = os.path.basename(os.path.splitext(filename)[0])
            if name not in KNOWN_LSSTTEXMF_BIB_NAMES:
                logger.warning('%r is not a known lsst-texmf bib file',
                               name)
                continue
            bibtex_names.append(name)

    # names of bibtex files not in cache
    uncached_names = [name for name in bibtex_names
                      if name not in _LSSTTEXMF_BIB_CACHE]
    if len(uncached_names) > 0:
        # Download bibtex and put into the cache
        loop = asyncio.get_event_loop()
        future = asyncio.ensure_future(_download_lsst_bibtex(uncached_names))
        loop.run_until_complete(future)
        for name, text in zip(bibtex_names, future.result()):
            _LSSTTEXMF_BIB_CACHE[name] = text

    return {name: _LSSTTEXMF_BIB_CACHE[name] for name in bibtex_names}


def get_bibliography(lsst_bib_names=None, bibtex=None):
    """Make a pybtex BibliographyData instance from standard lsst-texmf
    bibliography files and user-supplied bibtex content.

    Parameters
    ----------
    lsst_bib_names : sequence of `str`, optional
        Names of lsst-texmf BibTeX files to include. For example:

        .. code-block:: python

           ['lsst', 'lsst-dm', 'refs', 'books', 'refs_ads']

        Default is `None`, which includes all lsst-texmf bibtex files.

    bibtex : `str`
        BibTeX source content not included in lsst-texmf. This can be content
        from a import ``local.bib`` file.

    Returns
    -------
    bibliography : `pybtex.database.BibliographyData`
        A pybtex bibliography database that includes all given sources:
        lsst-texmf bibliographies and ``bibtex``.
    """
    bibtex_data = get_lsst_bibtex(bibtex_filenames=lsst_bib_names)

    # Parse with pybtex into BibliographyData instances
    pybtex_data = [pybtex.database.parse_string(_bibtex, 'bibtex')
                   for _bibtex in bibtex_data.values()]

    # Also parse local bibtex content
    if bibtex is not None:
        pybtex_data.append(pybtex.database.parse_string(bibtex, 'bibtex'))

    # Merge BibliographyData
    bib = pybtex_data[0]
    if len(pybtex_data) > 1:
        for other_bib in pybtex_data[1:]:
            for key, entry in other_bib.entries.items():
                bib.add_entry(key, entry)

    return bib


def get_url_from_entry(entry):
    """Get a usable URL from a pybtex entry.

    Parameters
    ----------
    entry : `pybtex.database.Entry`
        A pybtex bibliography entry.

    Returns
    -------
    url : `str`
        Best available URL from the ``entry``.

    Raises
    ------
    NoEntryUrlError
        Raised when no URL can be made from the bibliography entry.

    Notes
    -----
    The order of priority is:

    1. ``url`` field
    2. ``ls.st`` URL from the handle for ``@docushare`` entries.
    3. ``adsurl``
    4. DOI
    """
    if 'url' in entry.fields:
        return entry.fields['url']
    elif entry.type.lower() == 'docushare':
        return 'https://ls.st/' + entry.fields['handle']
    elif 'adsurl' in entry.fields:
        return entry.fields['adsurl']
    elif 'doi' in entry.fields:
        return 'https://doi.org/' + entry.fields['doi']
    else:
        raise NoEntryUrlError()


class NoEntryUrlError(RuntimeError):
    """Raised when a URL cannot be resolved from a bib entry."""


def get_authoryear_from_entry(entry, paren=False):
    """Get and format author-year text from a pybtex entry to emulate
    natbib citations.

    Parameters
    ----------
    entry : `pybtex.database.Entry`
        A pybtex bibliography entry.
    parens : `bool`, optional
        Whether to add parentheses around the year. Default is `False`.

    Returns
    -------
    authoryear : `str`
        The author-year citation text.
    """
    def _format_last(person):
        """Reformat a pybtex Person into a last name.

        Joins all parts of a last name and strips "{}" wrappers.
        """
        return ' '.join([n.strip('{}') for n in person.last_names])

    if len(entry.persons['author']) > 0:
        # Grab author list
        persons = entry.persons['author']
    elif len(entry.persons['editor']) > 0:
        # Grab editor list
        persons = entry.persons['editor']
    else:
        raise AuthorYearError

    try:
        year = entry.fields['year']
    except KeyError:
        raise AuthorYearError

    if paren and len(persons) == 1:
        template = '{author} ({year})'
        return template.format(author=_format_last(persons[0]),
                               year=year)
    elif not paren and len(persons) == 1:
        template = '{author} {year}'
        return template.format(author=_format_last(persons[0]),
                               year=year)
    elif paren and len(persons) == 2:
        template = '{author1} and {author2} ({year})'
        return template.format(author1=_format_last(persons[0]),
                               author2=_format_last(persons[1]),
                               year=year)
    elif not paren and len(persons) == 2:
        template = '{author1} and {author2} {year}'
        return template.format(author1=_format_last(persons[0]),
                               author2=_format_last(persons[1]),
                               year=year)
    elif not paren and len(persons) > 2:
        template = '{author} et al {year}'
        return template.format(author=_format_last(persons[0]),
                               year=year)
    elif paren and len(persons) > 2:
        template = '{author} et al ({year})'
        return template.format(author=_format_last(persons[0]),
                               year=year)


class AuthorYearError(RuntimeError):
    """Raised when an author-year citation cannot be made from a bib entry."""



"""Tests for lsstprojectmeta.tex.scraper.
"""

from lsstprojectmeta.tex import scraper


def test_get_def_macros():
    sample = r"\def \name {content}"
    macros = scraper.get_def_macros(sample)

    assert r'\name' in macros
    assert macros[r'\name'] == 'content'


def test_get_def_macros_LDM_503():
    sample = (r'\documentclass[DM,STP,toc]{lsstdoc}' + '\n'
              '%set the WP number or product here for the requirements\n'
              r'\def\product{Data Management}' + '\n'
              r'\def\cycle{S17}' + '\n')
    macros = scraper.get_def_macros(sample)

    assert macros[r'\product'] == 'Data Management'
    assert macros[r'\cycle'] == 'S17'


def test_get_newcommand_macros():
    sample = r"\newcommand {\name} {content}"
    macros = scraper.get_newcommand_macros(sample)
    assert macros[r'\name'] == 'content'

    sample = r"\newcommand { \name } {content}"
    macros = scraper.get_newcommand_macros(sample)
    assert macros[r'\name'] == 'content'

    sample = r"\newcommand{\name}{content}"
    macros = scraper.get_newcommand_macros(sample)
    assert macros[r'\name'] == 'content'



"""Tests for lsstprojectmeta.lsstdocument.handles
"""

import pytest

from lsstprojectmeta.lsstdocument.handles import DOCUMENT_HANDLE_PATTERN


@pytest.mark.parametrize(
    'text,expected_series,expected_serial', [
        ('LDM-151', 'LDM', '151'),
        ('LSE-163', 'LSE', '163'),
        ('LPM-231', 'LPM', '231'),
        ('DMTR-51', 'DMTR', '51'),
        ('DMTN-001', 'DMTN', '001'),
        ('SQR-001', 'SQR', '001'),
        ('SMTN-001', 'SMTN', '001'),
        ('ldm-151', 'ldm', '151'),
        ('lse-163', 'lse', '163'),
        ('lpm-231', 'lpm', '231'),
        ('dmtr-51', 'dmtr', '51'),
        ('dmtn-001', 'dmtn', '001'),
        ('sqr-001', 'sqr', '001'),
        ('smtn-001', 'smtn', '001')
    ])
def test_document_handle_pattern(text, expected_series, expected_serial):
    match = DOCUMENT_HANDLE_PATTERN.match(text)
    assert match.group('series') == expected_series
    assert match.group('serial') == expected_serial



"""Test LsstLatexDoc using sample data from LDM-nnn.tex.
"""

import datetime
import os
import pytest
from lsstprojectmeta.tex.lsstdoc import LsstLatexDoc

TITLE = "Title of document"

SHORT_TITLE = "Short title"

HTML_TITLE = "Title of document\n"

HTML_SHORT_TITLE = "Short title\n"

PLAIN_TITLE = "Title of document\n"

PLAIN_SHORT_TITLE = "Short title\n"

AUTHORS = ['A. Author', 'B. Author', 'C. Author']

HTML_AUTHORS = ['A. Author', 'B. Author', 'C. Author']

PLAIN_AUTHORS = ['A. Author', 'B. Author', 'C. Author']

ABSTRACT = (
    "This document demonstrates how to use the LSST \\LaTeX\\ "
    "class files to make Data Management\ndocuments. Build this "
    "document in the normal way, making sure that the class "
    "file is\navailable in the \\LaTeX\\ load path."
)

HTML_ABSTRACT = (
    '<p>This document demonstrates how to use the LSST LaTeX\xa0class files '
    'to make Data Management documents. Build this document in the normal '
    'way, making sure that the class file is available in the LaTeX\xa0load '
    'path.</p>\n'
)

PLAIN_ABSTRACT = (
    'This document demonstrates how to use the LSST LaTeX\xa0class files '
    'to make Data Management documents. Build this document in the normal '
    'way, making sure that the class file is available in the LaTeX\xa0load '
    'path.\n'
)

IS_DRAFT = True

HANDLE = 'LDM-nnn'

SERIES = 'LDM'

SERIAL = 'nnn'

ATTRIBUTES = [
    ('title', TITLE),
    ('short_title', SHORT_TITLE),
    ('html_title', HTML_TITLE),
    ('html_short_title', HTML_SHORT_TITLE),
    ('plain_title', PLAIN_TITLE),
    ('plain_short_title', PLAIN_SHORT_TITLE),
    ('authors', AUTHORS),
    ('html_authors', HTML_AUTHORS),
    ('plain_authors', PLAIN_AUTHORS),
    ('abstract', ABSTRACT),
    ('html_abstract', HTML_ABSTRACT),
    ('plain_abstract', PLAIN_ABSTRACT),
    ('is_draft', IS_DRAFT),
    ('handle', HANDLE),
    ('series', SERIES),
    ('serial', SERIAL),
]

JSONLD = {
    '@context': [
        "https://raw.githubusercontent.com/codemeta/codemeta/2.0-rc/"
        "codemeta.jsonld",
        "http://schema.org"],
    '@type': ['Report', 'SoftwareSourceCode'],
    'language': 'TeX',
    'reportNumber': HANDLE,
    'name': PLAIN_TITLE,
    'description': PLAIN_ABSTRACT,
    'author': [{'@type': 'Person', 'name': author_name}
               for author_name in PLAIN_AUTHORS],
}


@pytest.fixture
def lsstdoc():
    tex_path = os.path.join(os.path.dirname(__file__), 'data', 'LDM-nnn.tex')
    return LsstLatexDoc.read(tex_path)


@pytest.mark.parametrize('attribute,expected', ATTRIBUTES)
def test_attribute(lsstdoc, attribute, expected):
    assert getattr(lsstdoc, attribute) == expected


def test_revision_date(lsstdoc):
    r"""LDM-nnn is a draft so it falls back to git (would anyway since
    ``\date`` is ``\today``).
    """
    assert isinstance(lsstdoc.revision_datetime, datetime.datetime)
    assert lsstdoc.revision_datetime_source == 'git'


def test_jsonld(lsstdoc):
    jsonld = lsstdoc.build_jsonld()
    for key, value in JSONLD.items():
        assert jsonld[key] == value
    assert jsonld['dateModified'] == lsstdoc.revision_datetime



"""Metadata extraction from lsstdoc LSST LaTeX documents."""

__all__ = ['LsstLatexDoc']

import datetime
import logging
import os

from pytz import timezone
import pytz

from .commandparser import LatexCommand
from ..pandoc.convert import convert_lsstdoc_tex
from .scraper import get_macros
from .normalizer import read_tex_file, replace_macros
from .lsstbib import get_bibliography, KNOWN_LSSTTEXMF_BIB_NAMES
from .citelink import CitationLinker
from ..git.timestamp import get_content_commit_date


class LsstLatexDoc(object):
    """An lsstdoc-class LaTeX document with metadata access.

    Parameters
    ----------
    tex_source : `str`
        LaTeX source for the main file of an lsstdoc LaTeX document.
    root_dir : `str`, optional
        Root directory of the LaTeX project. `None` is treated as the
        current working directory.
    """

    def __init__(self, tex_source, root_dir=None):
        super().__init__()
        self._logger = logging.getLogger(__name__)

        # The BibliographyData (parsed BibTeX) is only loaded when
        # the self.bib_db attributed is first accessed.
        self._bib_db = None

        self._tex = tex_source
        if root_dir is None:
            self._root_dir = ''
        else:
            self._root_dir = root_dir

    @classmethod
    def read(cls, root_tex_path):
        """Construct an `LsstLatexDoc` instance by reading and parsing the
        LaTeX source.

        Parameters
        ----------
        root_tex_path : `str`
            Path to the LaTeX source on the filesystem. For multi-file LaTeX
            projects this should be the path to the root document.

        Notes
        -----
        This method implements the following pipeline:

        1. `lsstprojectmeta.tex.normalizer.read_tex_file`
        2. `lsstprojectmeta.tex.scraper.get_macros`
        3. `lsstprojectmeta.tex.normalizer.replace_macros`

        Thus ``input`` and ``includes`` are resolved along with simple macros.
        """
        # Read and normalize the TeX source, replacing macros with content
        root_dir = os.path.dirname(root_tex_path)
        tex_source = read_tex_file(root_tex_path)
        tex_macros = get_macros(tex_source)
        tex_source = replace_macros(tex_source, tex_macros)
        return cls(tex_source, root_dir=root_dir)

    @property
    def plain_content(self):
        """Plain-text-formatted document content (`str`)."""
        return self.format_content(format='plain', mathjax=False, smart=True)

    @property
    def html_title(self):
        """HTML5-formatted document title (`str`)."""
        return self.format_title(format='html5', deparagraph=True,
                                 mathjax=False, smart=True)

    @property
    def plain_title(self):
        """Plain-text-formatted document title (`str`)."""
        return self.format_title(format='plain', deparagraph=True,
                                 mathjax=False, smart=True)

    @property
    def title(self):
        """LaTeX-formatted document title (`str`)."""
        if not hasattr(self, '_title'):
            self._parse_title()

        return self._title

    @property
    def html_short_title(self):
        """HTML5-formatted document short title (`str`)."""
        return self.format_short_title(format='html5', deparagraph=True,
                                       mathjax=False, smart=True)

    @property
    def plain_short_title(self):
        """Plaintext-formatted document short title (`str`)."""
        return self.format_short_title(format='plain', deparagraph=True,
                                       mathjax=False, smart=True)

    @property
    def short_title(self):
        """LaTeX-formatted document short title (`str`)."""
        if not hasattr(self, '_short_title'):
            self._parse_title()

        return self._short_title

    @property
    def html_authors(self):
        """HTML5-formatted authors (`list` of `str`)."""
        return self.format_authors(format='html5', deparagraph=True,
                                   mathjax=False, smart=True)

    @property
    def plain_authors(self):
        """Plaintext-formatted authors (`list` of `str`)."""
        return self.format_authors(format='plain', deparagraph=True,
                                   mathjax=False, smart=True)

    @property
    def authors(self):
        """LaTeX-formatted authors (`list` of `str`)."""
        if not hasattr(self, '_authors'):
            self._parse_author()

        return self._authors

    @property
    def html_abstract(self):
        """HTML5-formatted document abstract (`str`)."""
        return self.format_abstract(format='html5', deparagraph=False,
                                    mathjax=False, smart=True)

    @property
    def plain_abstract(self):
        """Plaintext-formatted document abstract (`str`)."""
        return self.format_abstract(format='plain', deparagraph=False,
                                    mathjax=False, smart=True)

    @property
    def abstract(self):
        """LaTeX-formatted abstract (`str`)."""
        if not hasattr(self, '_abstract'):
            self._parse_abstract()

        return self._abstract

    @property
    def handle(self):
        """LaTeX-formatted document handle (`str`)."""
        if not hasattr(self, '_handle'):
            self._parse_doc_ref()

        return self._handle

    @property
    def series(self):
        """Document series identifier (`str`)."""
        if not hasattr(self, '_series'):
            self._parse_doc_ref()

        return self._series

    @property
    def serial(self):
        """Document serial number within series (`str`)."""
        if not hasattr(self, '_serial'):
            self._parse_doc_ref()

        return self._serial

    @property
    def is_draft(self):
        """Document is a draft if ``'lsstdoc'`` is included in the
        documentclass options (`bool`).
        """
        if not hasattr(self, '_document_options'):
            self._parse_documentclass()

        if 'lsstdraft' in self._document_options:
            return True
        else:
            return False

    @property
    def revision_datetime(self):
        """Current revision date of the document (`datetime.datetime`).

        The `revision_datetime_source` describes how the revision date
        is computed.

        This ``revision datetime`` is cached the first time you access it. This
        means that a datetime computed via ``now`` or ``git`` will not change
        during the lifetime of an `LsstLatexDoc` object.
        """
        if not hasattr(self, '_datetime'):
            self._parse_revision_date()

        return self._datetime

    @property
    def revision_datetime_source(self):
        r"""Data source for the `revision_datetime` attribute (`str`).

        Possible string values are:

        - ``'tex'``: The document revision date is defined in the ``\date``
          command. ``YYYY-MM-DD`` dates are converted to UTC datetimes by
          assuming the document is released at the beginning of the day in the
          ``US/Pacific`` timezone. Note: the ``\date`` command is ignored
          for draft documents (`is_draft` is `True`) so that drafts always
          fall back to ``'git'`` or ``'now'``.

        - ``'git'``: The latest Git commit's timestamp that affected document
          content. Content is considered any file with a ``tex``, ``bib``,
          ``pdf``, ``jpg``, or ``png`` extension. Git timestamps are used when
          the ``\date`` command is missing or can't be parsed.

        - ``'now'``: The current date and time. This source is used as a
          fallback when the LaTeX and Git-based methods of determining a
          document's date fail.

        The `revision datetime` is cached the first time you access it. This
        means that a datetime computed via ``now`` or ``git`` will not change
        during the lifetime of an `LsstLatexDoc` object.
        """
        if not hasattr(self, '_datetime'):
            self._parse_revision_date()

        return self._revision_datetime_source

    @property
    def bib_db(self):
        """Bibliography database referenced by the document
        (`pybtex.database.BibliographyData`).
        """
        if self._bib_db is None:
            # Load reference BibTeX into a pybtex BibliographyData
            self._load_bib_db()
        return self._bib_db

    def format_content(self, format='plain', mathjax=False,
                       smart=True, extra_args=None):
        """Get the document content in the specified markup format.

        Parameters
        ----------
        format : `str`, optional
            Output format (such as ``'html5'`` or ``'plain'``).
        mathjax : `bool`, optional
            Allow pandoc to use MathJax math markup.
        smart : `True`, optional
            Allow pandoc to create "smart" unicode punctuation.
        extra_args : `list`, optional
            Additional command line flags to pass to Pandoc. See
            `lsstprojectmeta.pandoc.convert.convert_text`.

        Returns
        -------
        output_text : `str`
            Converted content.
        """
        output_text = convert_lsstdoc_tex(
            self._tex, format,
            mathjax=mathjax,
            smart=smart,
            extra_args=extra_args)
        return output_text

    def format_title(self, format='html5', deparagraph=True, mathjax=False,
                     smart=True, extra_args=None):
        """Get the document title in the specified markup format.

        Parameters
        ----------
        format : `str`, optional
            Output format (such as ``'html5'`` or ``'plain'``).
        deparagraph : `bool`, optional
            Remove the paragraph tags from single paragraph content.
        mathjax : `bool`, optional
            Allow pandoc to use MathJax math markup.
        smart : `True`, optional
            Allow pandoc to create "smart" unicode punctuation.
        extra_args : `list`, optional
            Additional command line flags to pass to Pandoc. See
            `lsstprojectmeta.pandoc.convert.convert_text`.

        Returns
        -------
        output_text : `str`
            Converted content or `None` if the title is not available in
            the document.
        """
        if self.title is None:
            return None

        output_text = convert_lsstdoc_tex(
            self.title, format,
            deparagraph=deparagraph,
            mathjax=mathjax,
            smart=smart,
            extra_args=extra_args)
        return output_text

    def format_short_title(self, format='html5', deparagraph=True,
                           mathjax=False, smart=True, extra_args=None):
        """Get the document short title in the specified markup format.

        Parameters
        ----------
        format : `str`, optional
            Output format (such as ``'html5'`` or ``'plain'``).
        deparagraph : `bool`, optional
            Remove the paragraph tags from single paragraph content.
        mathjax : `bool`, optional
            Allow pandoc to use MathJax math markup.
        smart : `True`, optional
            Allow pandoc to create "smart" unicode punctuation.
        extra_args : `list`, optional
            Additional command line flags to pass to Pandoc. See
            `lsstprojectmeta.pandoc.convert.convert_text`.

        Returns
        -------
        output_text : `str`
            Converted content or `None` if the short title is not available in
            the document.
        """
        if self.short_title is None:
            return None

        output_text = convert_lsstdoc_tex(
            self.short_title, 'html5',
            deparagraph=deparagraph,
            mathjax=mathjax,
            smart=smart,
            extra_args=extra_args)
        return output_text

    def format_abstract(self, format='html5', deparagraph=False, mathjax=False,
                        smart=True, extra_args=None):
        """Get the document abstract in the specified markup format.

        Parameters
        ----------
        format : `str`, optional
            Output format (such as ``'html5'`` or ``'plain'``).
        deparagraph : `bool`, optional
            Remove the paragraph tags from single paragraph content.
        mathjax : `bool`, optional
            Allow pandoc to use MathJax math markup.
        smart : `True`, optional
            Allow pandoc to create "smart" unicode punctuation.
        extra_args : `list`, optional
            Additional command line flags to pass to Pandoc. See
            `lsstprojectmeta.pandoc.convert.convert_text`.

        Returns
        -------
        output_text : `str`
            Converted content or `None` if the title is not available in
            the document.
        """
        if self.abstract is None:
            return None

        abstract_latex = self._prep_snippet_for_pandoc(self.abstract)

        output_text = convert_lsstdoc_tex(
            abstract_latex, format,
            deparagraph=deparagraph,
            mathjax=mathjax,
            smart=smart,
            extra_args=extra_args)
        return output_text

    def format_authors(self, format='html5', deparagraph=True, mathjax=False,
                       smart=True, extra_args=None):
        """Get the document authors in the specified markup format.

        Parameters
        ----------
        format : `str`, optional
            Output format (such as ``'html5'`` or ``'plain'``).
        deparagraph : `bool`, optional
            Remove the paragraph tags from single paragraph content.
        mathjax : `bool`, optional
            Allow pandoc to use MathJax math markup.
        smart : `True`, optional
            Allow pandoc to create "smart" unicode punctuation.
        extra_args : `list`, optional
            Additional command line flags to pass to Pandoc. See
            `lsstprojectmeta.pandoc.convert.convert_text`.

        Returns
        -------
        output_text : `list` of `str`
            Sequence of author names in the specified output markup format.
        """
        formatted_authors = []
        for latex_author in self.authors:
            formatted_author = convert_lsstdoc_tex(
                latex_author, format,
                deparagraph=deparagraph,
                mathjax=mathjax,
                smart=smart,
                extra_args=extra_args)
            # removes Pandoc's terminal newlines
            formatted_author = formatted_author.strip()
            formatted_authors.append(formatted_author)
        return formatted_authors

    def _parse_documentclass(self):
        """Parse documentclass options.

        Sets the the ``_document_options`` attribute.
        """
        command = LatexCommand(
            'documentclass',
            {'name': 'options', 'required': False, 'bracket': '['},
            {'name': 'class_name', 'required': True, 'bracket': '{'})
        try:
            parsed = next(command.parse(self._tex))
        except StopIteration:
            self._logger.warning('lsstdoc has no documentclass')
            self._document_options = []

        try:
            content = parsed['options']
            self._document_options = [opt.strip()
                                      for opt in content.split(',')]
        except KeyError:
            self._logger.warning('lsstdoc has no documentclass options')
            self._document_options = []

    def _parse_title(self):
        """Parse the title from TeX source.

        Sets these attributes:

        - ``_title``
        - ``_short_title``
        """
        command = LatexCommand(
            'title',
            {'name': 'short_title', 'required': False, 'bracket': '['},
            {'name': 'long_title', 'required': True, 'bracket': '{'})
        try:
            parsed = next(command.parse(self._tex))
        except StopIteration:
            self._logger.warning('lsstdoc has no title')
            self._title = None
            self._short_title = None

        self._title = parsed['long_title']

        try:
            self._short_title = parsed['short_title']
        except KeyError:
            self._logger.warning('lsstdoc has no short title')
            self._short_title = None

    def _parse_doc_ref(self):
        """Parse the document handle.

        Sets the ``_series``, ``_serial``, and ``_handle`` attributes.
        """
        command = LatexCommand(
            'setDocRef',
            {'name': 'handle', 'required': True, 'bracket': '{'})
        try:
            parsed = next(command.parse(self._tex))
        except StopIteration:
            self._logger.warning('lsstdoc has no setDocRef')
            self._handle = None
            self._series = None
            self._serial = None
            return

        self._handle = parsed['handle']
        try:
            self._series, self._serial = self._handle.split('-', 1)
        except ValueError:
            self._logger.warning('lsstdoc handle cannot be parsed into '
                                 'series and serial: %r', self._handle)
            self._series = None
            self._serial = None

    def _parse_author(self):
        r"""Parse the author from TeX source.

        Sets the ``_authors`` attribute.

        Goal is to parse::

           \author{
           A.~Author,
           B.~Author,
           and
           C.~Author}

        Into::

           ['A. Author', 'B. Author', 'C. Author']
        """
        command = LatexCommand(
            'author',
            {'name': 'authors', 'required': True, 'bracket': '{'})
        try:
            parsed = next(command.parse(self._tex))
        except StopIteration:
            self._logger.warning('lsstdoc has no author')
            self._authors = []
            return

        try:
            content = parsed['authors']
        except KeyError:
            self._logger.warning('lsstdoc has no author')
            self._authors = []
            return

        # Clean content
        content = content.replace('\n', ' ')
        content = content.replace('~', ' ')
        content = content.strip()

        # Split content into list of individual authors
        authors = []
        for part in content.split(','):
            part = part.strip()
            for split_part in part.split('and '):
                split_part = split_part.strip()
                if len(split_part) > 0:
                    authors.append(split_part)
        self._authors = authors

    def _parse_abstract(self):
        """Parse the abstract from the TeX source.

        Sets the ``_abstract`` attribute.
        """
        command = LatexCommand(
            'setDocAbstract',
            {'name': 'abstract', 'required': True, 'bracket': '{'})
        try:
            parsed = next(command.parse(self._tex))
        except StopIteration:
            self._logger.warning('lsstdoc has no abstract')
            self._abstract = None
            return

        try:
            content = parsed['abstract']
        except KeyError:
            self._logger.warning('lsstdoc has no abstract')
            self._abstract = None
            return

        content = content.strip()
        self._abstract = content

    def _prep_snippet_for_pandoc(self, latex_text):
        """Process a LaTeX snippet of content for better transformation
        with pandoc.

        Currently runs the CitationLinker to convert BibTeX citations to
        href links.
        """
        replace_cite = CitationLinker(self.bib_db)
        latex_text = replace_cite(latex_text)
        return latex_text

    def _load_bib_db(self):
        r"""Load the BibTeX bibliography referenced by the document.

        This method triggered by the `bib_db` attribute and populates the
        `_bib_db` private attribute.

        The ``\bibliography`` command is parsed to identify the bibliographies
        referenced by the document.
        """
        # Get the names of custom bibtex files by parsing the
        # \bibliography command and filtering out the default lsstdoc
        # bibliographies.
        command = LatexCommand(
            'bibliography',
            {'name': 'bib_names', 'required': True, 'bracket': '{'})
        try:
            parsed = next(command.parse(self._tex))
            bib_names = [n.strip() for n in parsed['bib_names'].split(',')]
        except StopIteration:
            self._logger.warning('lsstdoc has no bibliography command')
            bib_names = []
        custom_bib_names = [n for n in bib_names
                            if n not in KNOWN_LSSTTEXMF_BIB_NAMES]

        # Read custom bibliographies.
        custom_bibs = []
        for custom_bib_name in custom_bib_names:
            custom_bib_path = os.path.join(
                os.path.join(self._root_dir),
                custom_bib_name + '.bib'
            )
            if not os.path.exists(custom_bib_path):
                self._logger.warning('Could not find bibliography %r',
                                     custom_bib_path)
                continue
            with open(custom_bib_path, 'r') as file_handle:
                custom_bibs.append(file_handle.read())
        if len(custom_bibs) > 0:
            custom_bibtex = '\n\n'.join(custom_bibs)
        else:
            custom_bibtex = None

        # Get the combined pybtex bibliography
        db = get_bibliography(bibtex=custom_bibtex)

        self._bib_db = db

    def _parse_revision_date(self):
        r"""Parse the ``\date`` command, falling back to getting the
        most recent Git commit date and the current datetime.

        Result is available from the `revision_datetime` attribute.
        """
        doc_datetime = None

        # First try to parse the \date command in the latex.
        # \date is ignored for draft documents.
        if not self.is_draft:
            date_command = LatexCommand(
                'date',
                {'name': 'content', 'required': True, 'bracket': '{'})
            try:
                parsed = next(date_command.parse(self._tex))
                command_content = parsed['content'].strip()
            except StopIteration:
                command_content = None
                self._logger.warning('lsstdoc has no date command')

            # Try to parse a date from the \date command
            if command_content is not None and command_content != r'\today':
                try:
                    doc_datetime = datetime.datetime.strptime(command_content,
                                                              '%Y-%m-%d')
                    # Assume LSST project time (Pacific)
                    project_tz = timezone('US/Pacific')
                    localized_datetime = project_tz.localize(doc_datetime)
                    # Normalize to UTC
                    doc_datetime = localized_datetime.astimezone(pytz.utc)

                    self._revision_datetime_source = 'tex'
                except ValueError:
                    self._logger.warning('Could not parse a datetime from '
                                         'lsstdoc date command: %r',
                                         command_content)

        # Fallback to getting the datetime from Git
        if doc_datetime is None:
            content_extensions = ('tex', 'bib', 'pdf', 'png', 'jpg')
            try:
                doc_datetime = get_content_commit_date(
                    content_extensions,
                    root_dir=self._root_dir)
                self._revision_datetime_source = 'git'
            except RuntimeError:
                self._logger.warning('Could not get a datetime from the Git '
                                     'repository at %r',
                                     self._root_dir)

        # Final fallback to the current datetime
        if doc_datetime is None:
            doc_datetime = pytz.utc.localize(datetime.datetime.now())
            self._revision_datetime_source = 'now'

        self._datetime = doc_datetime

    def build_jsonld(self, url=None, code_url=None, ci_url=None,
                     readme_url=None, license_id=None):
        """Create a JSON-LD representation of this LSST LaTeX document.

        Parameters
        ----------
        url : `str`, optional
            URL where this document is published to the web. Prefer
            the LSST the Docs URL if possible.
            Example: ``'https://ldm-151.lsst.io'``.
        code_url : `str`, optional
            Path the the document's repository, typically on GitHub.
            Example: ``'https://github.com/lsst/LDM-151'``.
        ci_url : `str`, optional
            Path to the continuous integration service dashboard for this
            document's repository.
            Example: ``'https://travis-ci.org/lsst/LDM-151'``.
        readme_url : `str`, optional
            URL to the document repository's README file. Example:
            ``https://raw.githubusercontent.com/lsst/LDM-151/master/README.rst``.
        license_id : `str`, optional
            License identifier, if known. The identifier should be from the
            listing at https://spdx.org/licenses/. Example: ``CC-BY-4.0``.

        Returns
        -------
        jsonld : `dict`
            JSON-LD-formatted dictionary.
        """
        jsonld = {
            '@context': [
                "https://raw.githubusercontent.com/codemeta/codemeta/2.0-rc/"
                "codemeta.jsonld",
                "http://schema.org"],
            '@type': ['Report', 'SoftwareSourceCode'],
            'language': 'TeX',
            'reportNumber': self.handle,
            'name': self.plain_title,
            'description': self.plain_abstract,
            'author': [{'@type': 'Person', 'name': author_name}
                       for author_name in self.plain_authors],
            # This is a datetime.datetime; not a string. If writing to a file,
            # Need to convert this to a ISO 8601 string.
            'dateModified': self.revision_datetime
        }

        try:
            jsonld['articleBody'] = self.plain_content
            jsonld['fileFormat'] = 'text/plain'  # MIME type of articleBody
        except RuntimeError:
            # raised by pypandoc when it can't convert the tex document
            self._logger.exception('Could not convert latex body to plain '
                                   'text for articleBody.')
            self._logger.warning('Falling back to tex source for articleBody')
            jsonld['articleBody'] = self._tex
            jsonld['fileFormat'] = 'text/plain'  # no mimetype for LaTeX?

        if url is not None:
            jsonld['@id'] = url
            jsonld['url'] = url
        else:
            # Fallback to using the document handle as the ID. This isn't
            # entirely ideal from a linked data perspective.
            jsonld['@id'] = self.handle

        if code_url is not None:
            jsonld['codeRepository'] = code_url

        if ci_url is not None:
            jsonld['contIntegration'] = ci_url

        if readme_url is not None:
            jsonld['readme'] = readme_url

        if license_id is not None:
            jsonld['license_id'] = None

        return jsonld



"""Test LsstLatexDoc using sample data from DMTN-044.tex.
"""

import os
import pytest
from lsstprojectmeta.tex.lsstdoc import LsstLatexDoc

TITLE = "LSST DM Software Release Considerations"

SHORT_TITLE = "DM Software Releases"

HTML_TITLE = "LSST DM Software Release Considerations\n"

HTML_SHORT_TITLE = "DM Software Releases\n"

PLAIN_TITLE = "LSST DM Software Release Considerations\n"

PLAIN_SHORT_TITLE = "DM Software Releases\n"

AUTHORS = ["John D. Swinbank"]

HTML_AUTHORS = ["John D. Swinbank"]

PLAIN_AUTHORS = ["John D. Swinbank"]

ABSTRACT = (
    "This attempts to summarise the debate around, and suggest a path "
    "forward, for\nLSST software releases. Although some recommendations "
    "are made, they are\nintended to serve as the basis of discussion, "
    "rather than as a complete\nsolution.\n\nThis material is based on "
    "discussions with several team members over a\nconsiderable period. "
    "Errors are to be expected; apologies are extended;\ncorrections are "
    "welcome."
)

HTML_ABSTRACT = (
    "<p>This attempts to summarise the debate around, and suggest a path "
    "forward, for LSST software releases. Although some recommendations are "
    "made, they are intended to serve as the basis of discussion, rather "
    "than as a complete solution.</p>"
    "<p>This material is based on discussions with several team members over "
    "a considerable period. Errors are to be expected; apologies are "
    "extended; corrections are welcome.</p>\n"
)

PLAIN_ABSTRACT = (
    "This attempts to summarise the debate around, and suggest a path "
    "forward, for LSST software releases. Although some recommendations are "
    "made, they are intended to serve as the basis of discussion, rather "
    "than as a complete solution.\n\n"
    "This material is based on discussions with several team members over "
    "a considerable period. Errors are to be expected; apologies are "
    "extended; corrections are welcome.\n"
)

IS_DRAFT = False

HANDLE = 'DMTN-044'

SERIES = 'DMTN'

SERIAL = '044'

ATTRIBUTES = [
    ('title', TITLE),
    ('short_title', SHORT_TITLE),
    ('html_title', HTML_TITLE),
    ('plain_title', PLAIN_TITLE),
    ('html_short_title', HTML_SHORT_TITLE),
    ('plain_short_title', PLAIN_SHORT_TITLE),
    ('authors', AUTHORS),
    ('html_authors', HTML_AUTHORS),
    ('plain_authors', PLAIN_AUTHORS),
    ('abstract', ABSTRACT),
    ('html_abstract', HTML_ABSTRACT),
    ('plain_abstract', PLAIN_ABSTRACT),
    ('is_draft', IS_DRAFT),
    ('handle', HANDLE),
    ('series', SERIES),
    ('serial', SERIAL),
]

JSONLD = {
    '@context': [
        "https://raw.githubusercontent.com/codemeta/codemeta/2.0-rc/"
        "codemeta.jsonld",
        "http://schema.org"],
    '@type': ['Report', 'SoftwareSourceCode'],
    'language': 'TeX',
    'reportNumber': HANDLE,
    'name': PLAIN_TITLE,
    'description': PLAIN_ABSTRACT,
    'author': [{'@type': 'Person', 'name': author_name}
               for author_name in PLAIN_AUTHORS],
}


@pytest.fixture
def lsstdoc():
    tex_path = os.path.join(os.path.dirname(__file__),
                            'data',
                            'DMTN-044',
                            'DMTN-044.tex')
    return LsstLatexDoc.read(tex_path)


@pytest.mark.parametrize('attribute,expected', ATTRIBUTES)
def test_attribute(lsstdoc, attribute, expected):
    assert getattr(lsstdoc, attribute) == expected


def test_revision_date(lsstdoc):
    r"""DMTN-044 sets a 2017-06-26 value for \setDocDate, which is deprecated
    and ignored (so it falls back to Git).
    """
    assert lsstdoc.revision_datetime_source == 'git'


def test_jsonld(lsstdoc):
    jsonld = lsstdoc.build_jsonld()
    for key, value in JSONLD.items():
        assert jsonld[key] == value
    assert jsonld['dateModified'] == lsstdoc.revision_datetime



"""Utilities for extracting commit timestamps from Git repositories.
"""

__all__ = ('read_git_commit_timestamp', 'read_git_commit_timestamp_for_file',
           'get_content_commit_date')

import itertools
import logging
import os

import git


def read_git_commit_timestamp(repo_path=None, repo=None):
    """Obtain the timestamp from the current head commit of a Git repository.

    Parameters
    ----------
    repo_path : `str`, optional
        Path to the Git repository. Leave as `None` to use the current working
        directory.

    Returns
    -------
    commit_timestamp : `datetime.datetime`
        The datetime of the head commit.
    """
    if repo is None:
        repo = git.repo.base.Repo(path=repo_path,
                                  search_parent_directories=True)
    head_commit = repo.head.commit
    return head_commit.committed_datetime


def read_git_commit_timestamp_for_file(filepath, repo_path=None, repo=None):
    """Obtain the timestamp for the most recent commit to a given file in a
    Git repository.

    Parameters
    ----------
    filepath : `str`
        Absolute or repository-relative path for a file.
    repo_path : `str`, optional
        Path to the Git repository. Leave as `None` to use the current working
        directory or if a ``repo`` argument is provided.
    repo : `git.Repo`, optional
        A `git.Repo` instance.

    Returns
    -------
    commit_timestamp : `datetime.datetime`
        The datetime of the most recent commit to the given file.

    Raises
    ------
    IOError
        Raised if the ``filepath`` does not exist in the Git repository.
    """
    logger = logging.getLogger(__name__)

    if repo is None:
        repo = git.repo.base.Repo(path=repo_path,
                                  search_parent_directories=True)
    repo_path = repo.working_tree_dir

    head_commit = repo.head.commit

    # filepath relative to the repo path
    logger.debug('Using Git repo at %r', repo_path)
    filepath = os.path.relpath(
        os.path.abspath(filepath),
        start=repo_path)
    logger.debug('Repo-relative filepath is %r', filepath)

    # Most recent commit datetime of the given file.
    # Don't use head_commit.iter_parents because then it skips the
    # commit of a file that's added but never modified.
    for commit in head_commit.iter_items(repo,
                                         head_commit,
                                         [filepath],
                                         skip=0):
        return commit.committed_datetime

    # Only get here if git could not find the file path in the history
    raise IOError('File {} not found'.format(filepath))


def get_content_commit_date(extensions, acceptance_callback=None,
                            root_dir='.'):
    """Get the datetime for the most recent commit to a project that
    affected certain types of content.

    Parameters
    ----------
    extensions : sequence of 'str'
        Extensions of files to consider in getting the most recent commit
        date. For example, ``('rst', 'svg', 'png')`` are content extensions
        for a Sphinx project. **Extension comparision is case sensitive.** add
        uppercase variants to match uppercase extensions.
    acceptance_callback : callable
        Callable function whose sole argument is a file path, and returns
        `True` or `False` depending on whether the file's commit date should
        be considered or not. This callback is only run on files that are
        included by ``extensions``. Thus this callback is a way to exclude
        specific files that would otherwise be included by their extension.
    root_dir : 'str`, optional
        Only content contained within this root directory is considered.
        This directory must be, or be contained by, a Git repository. This is
        the current working directory by default.

    Returns
    -------
    commit_date : `datetime.datetime`
        Datetime of the most recent content commit.

    Raises
    ------
    RuntimeError
        Raised if no content files are found.
    """
    logger = logging.getLogger(__name__)

    def _null_callback(_):
        return True

    if acceptance_callback is None:
        acceptance_callback = _null_callback

    # Cache the repo object for each query
    root_dir = os.path.abspath(root_dir)
    repo = git.repo.base.Repo(path=root_dir, search_parent_directories=True)

    # Iterate over all files with all file extensions, looking for the
    # newest commit datetime.
    newest_datetime = None
    iters = [_iter_filepaths_with_extension(ext, root_dir=root_dir)
             for ext in extensions]
    for content_path in itertools.chain(*iters):
        content_path = os.path.abspath(os.path.join(root_dir, content_path))

        if acceptance_callback(content_path):
            logger.debug('Found content path %r', content_path)
            try:
                commit_datetime = read_git_commit_timestamp_for_file(
                    content_path, repo=repo)
                logger.debug('Commit timestamp of %r is %s',
                             content_path, commit_datetime)
            except IOError:
                logger.warning(
                    'Count not get commit for %r, skipping',
                    content_path)
                continue

            if not newest_datetime or commit_datetime > newest_datetime:
                # Seed initial newest_datetime
                # or set a newer newest_datetime
                newest_datetime = commit_datetime
                logger.debug('Newest commit timestamp is %s', newest_datetime)

        logger.debug('Final commit timestamp is %s', newest_datetime)

    if newest_datetime is None:
        raise RuntimeError('No content files found in {}'.format(root_dir))

    return newest_datetime


def _iter_filepaths_with_extension(extname, root_dir='.'):
    """Iterative over relative filepaths of files in a directory, and
    sub-directories, with the given extension.

    Parameters
    ----------
    extname : `str`
        Extension name (such as 'txt' or 'rst'). Extension comparison is
        case sensitive.
    root_dir : 'str`, optional
        Root directory. Current working directory by default.

    Yields
    ------
    filepath : `str`
        File path, relative to ``root_dir``, with the given extension.
    """
    # needed for comparison with os.path.splitext
    if not extname.startswith('.'):
        extname = '.' + extname

    root_dir = os.path.abspath(root_dir)

    for dirname, sub_dirnames, filenames in os.walk(root_dir):
        for filename in filenames:
            if os.path.splitext(filename)[-1] == extname:
                full_filename = os.path.join(dirname, filename)
                rel_filepath = os.path.relpath(full_filename, start=root_dir)
                yield rel_filepath



__all__ = ['CitationLinker']

from .commandparser import LatexCommand

from .lsstbib import (
    get_url_from_entry, NoEntryUrlError, get_authoryear_from_entry,
    AuthorYearError)


class CitationLinker(object):
    r"""LaTeX source processor that converts citation commands to ``\href``
    commands.

    This processing is useful for decoupling BibTeX from extracted TeX source
    snippets, like abstracts, that are intended to be converted into another
    markup language by pandoc.

    Parameters
    ----------
    bibtex_database : `pybtex.database.BibliographyData`
        A pybtex bibliography. Use `lsstprojectmeta.lsstbib.get_bibliography`
        to get this.
    """

    def __init__(self, bibtex_database):
        super().__init__()
        self._db = bibtex_database

        # Register and build Linker classes. They all share the same API.
        self._linker_classes = [CitedsLinker, CitedspLinker]
        self._linkers = [cls(self._db) for cls in self._linker_classes]

    def __call__(self, tex_source):
        r"""Convert citations in LaTeX source to Hyperref links.

        Parameters
        ----------
        tex_source : `str`
            LaTeX document source.

        Returns
        -------
        processed_tex : `str`
            LaTeX document source with all citation commands converted to
            ``\hyperref`` commands.
        """
        for linker in self._linkers:
            tex_source = linker(tex_source)
        return tex_source


class BaseCommandLinker(object):
    """Baseclass for citation linkers that process specific types of
    LaTeX commands.
    """

    def __call__(self, tex_source):
        r"""Convert commands of type ``command`` in LaTeX source to Hyperref
        links.

        Parameters
        ----------
        tex_source : `str`
            LaTeX document source.

        Returns
        -------
        processed_tex : `str`
            LaTeX document source with commands of type ``command`` to
            ``\hyperref`` commands.
        """
        while True:
            try:
                parsed = next(self.command.parse(tex_source))
            except StopIteration:
                break
            tex_source = self._replace_command(tex_source, parsed)
        return tex_source


class CitedsLinker(BaseCommandLinker):
    r"""Replace a ``\citeds`` citation with ``\href`` command.

    Examples
    --------
    >>> replace_citeds = CitedsLinker()
    >>> print(replace_citeds(r'\citeds{LDM-151}'))
    \href{https://ls.st/LDM-151}{LDM-151}

    Variant with defined title text:

    >>> print(replace_citeds(r'\citeds[Pipelines Design]{LDM-151}'))
    \href{https://ls.st/LDM-151}{Pipelines Design}
    """

    parens = ('', '')
    """Parentheses characters."""

    command_name = 'citeds'
    """Name of LaTeX command."""

    def __init__(self, bibtex_database=None):
        super().__init__()
        self._db = bibtex_database
        self.command = LatexCommand(
            self.command_name,
            {'bracket': '[', 'required': False, 'name': 'title'},
            {'bracket': '{', 'required': True, 'name': 'citekey'}
        )
        self.template = (
            self.parens[0] + r'\href{{{url}}}{{{content}}}' + self.parens[1]
        )

    def _replace_command(self, tex_source, parsed):
        if 'title' in parsed:
            content = parsed['title']
        else:
            # The document handle
            # Could get this from BibTeX
            content = parsed['citekey']

        url = 'https://ls.st/{citekey}'.format(citekey=parsed['citekey'])

        href_command = self.template.format(url=url, content=content)
        tex_source = tex_source.replace(
            parsed.command_source,
            href_command)

        return tex_source


class CitedspLinker(CitedsLinker):
    r"""Replace a ``\citedsp`` citation with ``\href`` command.

    Examples
    --------
    >>> replace_citedsp = CitedspLinker()
    >>> print(replace_citedsp(r'\citedsp{LDM-151}'))
    [\href{https://ls.st/LDM-151}{LDM-151}]

    Variant with defined title text:

    >>> print(replace_citedsp(r'\citedsp[Pipelines Design]{LDM-151}'))
    [\href{https://ls.st/LDM-151}{Pipelines Design}]
    """

    parens = ('[', ']')
    """Parentheses characters."""

    command_name = 'citedsp'
    """Name of LaTeX command."""


class CitepLinker(BaseCommandLinker):
    r"""Replace a ``\citep`` citation with an ``\href`` command.

    Examples
    --------
    >>> from pybtex.database import parse_string
    >>> bibitem = r'''
    ... @ARTICLE{2001ApJ...550..212B,
    ...   author = {{Bell}, E.~F. and {de Jong}, R.~S.},
    ...   title = "{Stellar Mass-to-Light Ratios and the Tully-Fisher
    ...     Relation}",
    ...   journal = {\apj},
    ...   eprint = {astro-ph/0011493},
    ...   keywords = {ISM: Dust, Extinction, Galaxies: Evolution,
    ...     Galaxies: Kinematics and Dynamics, Galaxies: Spiral,
    ...     Galaxies: Stellar Content},
    ...   year = 2001,
    ...   month = mar,
    ...   volume = 550,
    ...   pages = {212-229},
    ...   doi = {10.1086/319728},
    ...   adsurl = {http://adsabs.harvard.edu/abs/2001ApJ...550..212B},
    ...   adsnote = {Provided by the SAO/NASA Astrophysics Data System}
    ... }
    ... '''
    >>> bib_db = parse_string(bibitem, 'bibtex')
    >>> link_citep = CitepLinker(bib_db)
    >>> sample_text = r"\citep{2001ApJ...550..212B}"
    >>> print(link_citep(sample_text))  # doctest: +ELLIPSIS
    [\href{http://adsabs.harvard.edu/abs/2001ApJ...}{Bell and de Jong 2001}]
    """

    def __init__(self, bibtex_database=None):
        super().__init__()
        self._db = bibtex_database
        self.command = LatexCommand(
            'citep',
            {'bracket': '[', 'required': False, 'name': 'prefix'},
            {'bracket': '[', 'required': False, 'name': 'suffix'},
            {'bracket': '{', 'required': True, 'name': 'citekeys'}
        )
        self.outer_template = "[{content}]"
        self.link_template = r"\href{{{url}}}{{{content}}}"

    def _replace_command(self, tex_source, parsed):
        cite_keys = [k.strip() for k in parsed['citekeys'].split(',')]

        hrefs = []
        for cite_key in cite_keys:
            if self._db and cite_key in self._db.entries:
                entry = self._db.entries[cite_key]
                try:
                    url = get_url_from_entry(entry)
                except NoEntryUrlError:
                    url = None

                try:
                    authoryear = get_authoryear_from_entry(entry, paren=False)
                except AuthorYearError:
                    authoryear = None

                if url is not None and authoryear is not None:
                    hrefs.append(self.link_template.format(url=url,
                                                           content=authoryear))
                elif url is None and authoryear is not None:
                    # No link in this case
                    hrefs.append(authoryear)
                elif url is not None and authoryear is None:
                    # Link with cite key
                    hrefs.append(self.link_template.format(url=url,
                                                           content=cite_key))
                else:
                    # Just show cite key
                    hrefs.append(cite_key)

        replacement = self.outer_template.format(content=', '.join(hrefs))
        tex_source = tex_source.replace(
            parsed.command_source,
            replacement)

        return tex_source



"""Test LsstLatexDoc using sample data from LSE-61.tex.
"""

import datetime
import os
import pytest
import pytz
from lsstprojectmeta.tex.lsstdoc import LsstLatexDoc

TITLE = "Data Management System (DMS) Requirements"

SHORT_TITLE = None

HTML_TITLE = "Data Management System (DMS) Requirements\n"

HTML_SHORT_TITLE = None

PLAIN_TITLE = "Data Management System (DMS) Requirements\n"

PLAIN_SHORT_TITLE = None

AUTHORS = [
    "Gregory Dubois-Felsmann", "Tim Jenness"
]

HTML_AUTHORS = [
    "Gregory Dubois-Felsmann", "Tim Jenness"
]

PLAIN_AUTHORS = [
    "Gregory Dubois-Felsmann", "Tim Jenness"
]

ABSTRACT = None

HTML_ABSTRACT = None

PLAIN_ABSTRACT = None

IS_DRAFT = False

HANDLE = 'LSE-61'

SERIES = 'LSE'

SERIAL = '61'

ATTRIBUTES = [
    ('title', TITLE),
    ('short_title', SHORT_TITLE),
    ('html_title', HTML_TITLE),
    ('html_short_title', HTML_SHORT_TITLE),
    ('plain_title', PLAIN_TITLE),
    ('plain_short_title', PLAIN_SHORT_TITLE),
    ('authors', AUTHORS),
    ('html_authors', HTML_AUTHORS),
    ('plain_authors', PLAIN_AUTHORS),
    ('abstract', ABSTRACT),
    ('html_abstract', HTML_ABSTRACT),
    ('plain_abstract', PLAIN_ABSTRACT),
    ('is_draft', IS_DRAFT),
    ('handle', HANDLE),
    ('series', SERIES),
    ('serial', SERIAL),
]

JSONLD = {
    '@context': [
        "https://raw.githubusercontent.com/codemeta/codemeta/2.0-rc/"
        "codemeta.jsonld",
        "http://schema.org"],
    '@type': ['Report', 'SoftwareSourceCode'],
    'language': 'TeX',
    'reportNumber': HANDLE,
    'name': PLAIN_TITLE,
    'description': PLAIN_ABSTRACT,
    'author': [{'@type': 'Person', 'name': author_name}
               for author_name in PLAIN_AUTHORS],
}


@pytest.fixture
def lsstdoc():
    tex_path = os.path.join(os.path.dirname(__file__),
                            'data',
                            'LSE-61',
                            'LSE-61.tex')
    return LsstLatexDoc.read(tex_path)


@pytest.mark.parametrize('attribute,expected', ATTRIBUTES)
def test_attribute(lsstdoc, attribute, expected):
    assert getattr(lsstdoc, attribute) == expected


def test_revision_date(lsstdoc):
    r"""LSE-61 uses a set value for \date."""
    expected_datetime = datetime.datetime(2017, 9, 12, 7, 0, tzinfo=pytz.utc)
    assert lsstdoc.revision_datetime == expected_datetime
    assert lsstdoc.revision_datetime_source == 'tex'


def test_jsonld(lsstdoc):
    jsonld = lsstdoc.build_jsonld()
    for key, value in JSONLD.items():
        assert jsonld[key] == value
    assert jsonld['dateModified'] == lsstdoc.revision_datetime



"""Tests for the lsstprojectmeta.jsonld module.
"""

import datetime
from lsstprojectmeta.jsonld import encode_jsonld, decode_jsonld


def test_datetime_encoding():
    """Test that a datetime.datetime can be encoded.
    """
    test_date = datetime.datetime(2018, 1, 1, hour=12, minute=0, second=0,
                                  tzinfo=datetime.timezone.utc)
    jsonld_doc = {'date': test_date}

    expected = '{"date": "2018-01-01T12:00:00Z"}'

    assert expected == encode_jsonld(jsonld_doc)


def test_datetime_encoding_no_tz():
    """Test that a datetime.datetime with no timezone can be encoded.
    """
    test_date = datetime.datetime(2018, 1, 1, hour=12, minute=0, second=0)
    jsonld_doc = {'date': test_date}

    expected = '{"date": "2018-01-01T12:00:00Z"}'

    assert expected == encode_jsonld(jsonld_doc)


def test_datetime_decoding():
    """Test that a datetime object can be decoded from JSON.
    """
    expected_date = datetime.datetime(2018, 1, 1, hour=12, minute=0, second=0,
                                      tzinfo=datetime.timezone.utc)
    expected = {'date': expected_date, 'alist': [1, 2, 3]}

    jsonld_text = '{"date": "2018-01-01T12:00:00Z", "alist": [1, 2, 3]}'

    json_doc = decode_jsonld(jsonld_text)

    assert json_doc['date'] == expected_date
    assert json_doc == expected



"""Functions for normalizing TeX source.
"""

__all__ = ['remove_comments', 'remove_trailing_whitespace', 'read_tex_file',
           'process_inputs', 'replace_macros']

import logging
import os
import re


# Regular expression for finding input or include commands
input_include_pattern = re.compile(
    r'\\(?P<command>input|include)'  # command name
    r'(([ ]*\{)|([ ]+))'
    r'(?P<filename>[\w/\-\.]+)'  # included filename
    r'[\}%\s]'  # closing whitespace or bracket
)
input_pattern = re.compile(
    r'\\input[ ]*?{*?(?P<filename>[\w/\-\.]+)[\}%\s]')
include_pattern = re.compile(
    r'\\include[ ]*?{*?(?P<filename>[\w/\-\.]+)[\}%\s]')


def remove_comments(tex_source):
    """Delete latex comments from TeX source.

    Parameters
    ----------
    tex_source : str
        TeX source content.

    Returns
    -------
    tex_source : str
        TeX source without comments.
    """
    # Expression via http://stackoverflow.com/a/13365453
    return re.sub(r'(?<!\\)%.*$', r'', tex_source, flags=re.M)


def remove_trailing_whitespace(tex_source):
    """Delete trailing whitespace from TeX source.

    Parameters
    ----------
    tex_source : str
        TeX source content.

    Returns
    -------
    tex_source : str
        TeX source without trailing whitespace.
    """
    # Delete any space or tab characters right before a new line
    return re.sub(r'[ \t]+$', '', tex_source, flags=re.M)


def read_tex_file(root_filepath, root_dir=None):
    r"""Read a TeX file, automatically processing and normalizing it
    (including other input files, removing comments, and deleting trailing
    whitespace).

    Parameters
    ----------
    root_filepath : `str`
        Filepath to a TeX file.
    root_dir : `str`
        Root directory of the TeX project. This only needs to be set when
        recursively reading in ``\input`` or ``\include`` files.

    Returns
    -------
    tex_source : `str`
        TeX source.
    """
    with open(root_filepath, 'r') as f:
        tex_source = f.read()

    if root_dir is None:
        root_dir = os.path.dirname(root_filepath)

    # Text processing pipline
    tex_source = remove_comments(tex_source)
    tex_source = remove_trailing_whitespace(tex_source)
    tex_source = process_inputs(tex_source, root_dir=root_dir)

    return tex_source


def process_inputs(tex_source, root_dir=None):
    r"""Insert referenced TeX file contents (from  ``\input`` and ``\include``
    commands) into the source.

    Parameters
    ----------
    tex_source : `str`
        TeX source where referenced source files will be found and inserted.
    root_dir : `str`, optional
        Name of the directory containing the TeX project's root file. Files
        referenced by TeX ``\input`` and ``\include`` commands are relative to
        this directory. If not set, the current working directory is assumed.

    Returns
    -------
    tex_source : `str`
        TeX source.

    See also
    --------
    `read_tex_file`
        Recommended API for reading a root TeX source file and inserting
        referenced files.
    """
    logger = logging.getLogger(__name__)

    def _sub_line(match):
        """Function to be used with re.sub to inline files for each match."""
        fname = match.group('filename')
        if not fname.endswith('.tex'):
            full_fname = ".".join((fname, 'tex'))
        else:
            full_fname = fname
        full_path = os.path.abspath(os.path.join(root_dir, full_fname))

        try:
            included_source = read_tex_file(full_path, root_dir=root_dir)
        except IOError:
            logger.error("Cannot open {0} for inclusion".format(full_path))
            raise
        else:
            return included_source

    tex_source = input_include_pattern.sub(_sub_line, tex_source)
    return tex_source


def replace_macros(tex_source, macros):
    r"""Replace macros in the TeX source with their content.

    Parameters
    ----------
    tex_source : `str`
        TeX source content.
    macros : `dict`
        Keys are macro names (including leading ``\``) and values are the
        content (as `str`) of the macros. See
        `lsstprojectmeta.tex.scraper.get_macros`.

    Returns
    -------
    tex_source : `str`
        TeX source with known macros replaced.

    Notes
    -----
    Macros with arguments are not supported.

    Examples
    --------
    >>> macros = {r'\handle': 'LDM-nnn'}
    >>> sample = r'This is document \handle.'
    >>> replace_macros(sample, macros)
    'This is document LDM-nnn.'

    Any trailing slash after the macro command is also replaced by this
    function.

    >>> macros = {r'\product': 'Data Management'}
    >>> sample = r'\title    [Test Plan]  { \product\ Test Plan}'
    >>> replace_macros(sample, macros)
    '\\title    [Test Plan]  { Data Management Test Plan}'
    """
    for macro_name, macro_content in macros.items():
        # '\\?' suffix matches an optional trailing '\' that might be used
        # for spacing.
        pattern = re.escape(macro_name) + r"\\?"
        # Wrap macro_content in lambda to avoid processing escapes
        tex_source = re.sub(pattern, lambda _: macro_content, tex_source)
    return tex_source



"""Utilities for parsing document handles and series information.
"""

__all__ = ('SERIES', 'DOCUMENT_HANDLE_PATTERN')

import re


SERIES = {
    'LPM': 'LSST Project Management',
    'LSE': 'LSST Systems Engineering',
    'LDM': 'LSST Data Management',
    'DMTR': 'LSST DM Test Report',
    'SQR': 'SQuaRE Technical Note',
    'DMTN': 'Data Management Technical Note',
    'SMTN': 'Simulations Technical Note',
    'PSTN': 'Project Science Team Technical Note',
    'SITCOMTN': 'Systems Integration, Testing, and Commissioning Technical '
                'Note',
}
"""Mapping between LSST document series (handle prefixes) and the title of the
series.
"""


DOCUMENT_HANDLE_PATTERN = re.compile(
    r'^(?P<series>' + '|'.join([h for h in SERIES]) + ')'
    r'-(?P<serial>\d+)',
    re.IGNORECASE)
"""Pattern that matches the handle of any LSST document.

Notes
-----
The pattern exposes two named groups in the match object:

- ``'series'``. The document series. For example, ``'LDM'``.
- ``'serial'``. The serial number, as a `str`. For example, ``'151'``.

Note that the pattern is **case insensitive.** If you input text is normalized
to lower case, it will still match, but the series will be in lower case.

Examples
--------

>>> m = DOCUMENT_HANDLE_PATTERN.match('LDM-151'.lower())
>>> m.group('series')
'LDM'
>>> m.group('serial')
'151'
"""



import logging

from pkg_resources import get_distribution, DistributionNotFound

__all__ = ('__version__',)


try:
    __version__ = get_distribution('lsst-projectmeta-kit').version
except DistributionNotFound:
    # Package is not installed
    __version__ = 'unknown'

# Allow users to attach their own handlers
logging.getLogger(__name__).addHandler(logging.NullHandler())



"""Tests for the lsstprojectmeta.tex.lsstbib module.
"""

from pybtex.database import BibliographyData, parse_string
import pytest

from lsstprojectmeta.tex.lsstbib import (
    get_lsst_bibtex, get_bibliography,
    get_url_from_entry, NoEntryUrlError,
    get_authoryear_from_entry, AuthorYearError)


def test_get_lsst_bibtex():
    bibtex = get_lsst_bibtex()
    assert 'lsst' in bibtex
    assert 'lsst-dm' in bibtex
    assert 'refs' in bibtex
    assert 'books' in bibtex
    assert 'refs_ads' in bibtex

    # Repeat (exercises cache)
    bibtex = get_lsst_bibtex()
    assert 'lsst' in bibtex
    assert 'lsst-dm' in bibtex
    assert 'refs' in bibtex
    assert 'books' in bibtex
    assert 'refs_ads' in bibtex


def test_get_bibliography():
    bib = get_bibliography()
    assert isinstance(bib, BibliographyData)

    # Key from 'lsst' bibtex file
    assert 'LDM-151' in bib.entries
    # Key from 'lsst-dm' bibtex file
    assert '2002SPIE.4844..225C' in bib.entries
    # Key from 'refs' bibtex file
    assert '1991.Spyak.OpticalEngineering' in bib.entries
    # Key from 'books' bibtex file
    assert 'Adass08' in bib.entries
    # Key from 'refs_ads' bibtex file
    assert '1727RSPT...35..637B' in bib.entries


def test_get_docushare_url_from_entry():
    bibtex = (
        '@DocuShare{LDM-151,\n'
        '    author =       {John D. Swinbank and others},\n'
        '    title =        "{Data Management Science Pipelines Design}",\n'
        '    year =         2017,\n'
        '    month =        may,\n'
        '    handle =       {LDM-151},\n'
        '}\n'
    )
    db = parse_string(bibtex, 'bibtex')
    entry = db.entries['LDM-151']
    assert get_url_from_entry(entry) == 'https://ls.st/LDM-151'


def test_get_ads_url_from_entry():
    bibtex = (
        r'@ARTICLE{2009MNRAS.400.1181Z,' + '\n'
        r' author = {{Zibetti}, S. and {Charlot}, S. and {Rix}, H.-W.},' + '\n'
        r' title = "{Resolved stellar mass maps of galaxies - I. Method and '
        r' implications for global mass estimates}",' + '\n'
        r' journal = {\mnras},' + '\n'
        r'archivePrefix = "arXiv",' + '\n'
        r' eprint = {0904.4252},' + '\n'
        r' primaryClass = "astro-ph.CO",' + '\n'
        r' keywords = {techniques: image processing , '
        r'techniques: photometric , galaxies: fundamental parameters , '
        r'galaxies: general , galaxies: photometry , '
        r'galaxies: stellar content},' + '\n'
        r'     year = 2009,' + '\n'
        r'    month = dec,' + '\n'
        r'   volume = 400,' + '\n'
        r'    pages = {1181-1198},' + '\n'
        r'      doi = {10.1111/j.1365-2966.2009.15528.x},' + '\n'
        r'adsurl = {http://adsabs.harvard.edu/abs/2009MNRAS.400.1181Z},' + '\n'
        r'adsnote = {Provided by the SAO/NASA Astrophysics Data System}' + '\n'
        r'}' + '\n'
    )
    db = parse_string(bibtex, 'bibtex')
    entry = db.entries['2009MNRAS.400.1181Z']
    expected = 'http://adsabs.harvard.edu/abs/2009MNRAS.400.1181Z'
    assert get_url_from_entry(entry) == expected


def test_get_doi_url_from_entry():
    bibtex = (
        '@ARTICLE{2009MNRAS.400.1181Z,\n'
        '  author = {{Zibetti}, S. and {Charlot}, S. and {Rix}, H.-W.},\n'
        '  title = "{Resolved stellar mass maps of galaxies - I. Method and '
        '  implications for global mass estimates}",\n'
        r'  journal = {\mnras},' + '\n'
        'archivePrefix = "arXiv",\n'
        '   eprint = {0904.4252},\n'
        ' primaryClass = "astro-ph.CO",\n'
        ' keywords = {techniques: image processing , '
        'techniques: photometric , galaxies: fundamental parameters , '
        'galaxies: general , galaxies: photometry , '
        'galaxies: stellar content},\n'
        '     year = 2009,\n'
        '    month = dec,\n'
        '   volume = 400,\n'
        '    pages = {1181-1198},\n'
        '      doi = {10.1111/j.1365-2966.2009.15528.x},\n'
        '  adsnote = {Provided by the SAO/NASA Astrophysics Data System}\n'
        '}\n'
    )
    db = parse_string(bibtex, 'bibtex')
    entry = db.entries['2009MNRAS.400.1181Z']
    expected = 'https://doi.org/10.1111/j.1365-2966.2009.15528.x'
    assert get_url_from_entry(entry) == expected


def test_get_url_from_entry():
    bibtex = (
        '@ARTICLE{2009MNRAS.400.1181Z,\n'
        '  author = {{Zibetti}, S. and {Charlot}, S. and {Rix}, H.-W.},\n'
        '  title = "{Resolved stellar mass maps of galaxies - I. Method and '
        '  implications for global mass estimates}",\n'
        r'  journal = {\mnras},' + '\n'
        'archivePrefix = "arXiv",\n'
        '   eprint = {0904.4252},\n'
        ' primaryClass = "astro-ph.CO",\n'
        ' keywords = {techniques: image processing , '
        'techniques: photometric , galaxies: fundamental parameters , '
        'galaxies: general , galaxies: photometry , '
        'galaxies: stellar content},\n'
        '     year = 2009,\n'
        '    month = dec,\n'
        '   volume = 400,\n'
        '    pages = {1181-1198},\n'
        '   url = {http://adsabs.harvard.edu/abs/2009MNRAS.400.1181Z},\n'
        '  adsnote = {Provided by the SAO/NASA Astrophysics Data System}\n'
        '}\n'
    )
    db = parse_string(bibtex, 'bibtex')
    entry = db.entries['2009MNRAS.400.1181Z']
    expected = 'http://adsabs.harvard.edu/abs/2009MNRAS.400.1181Z'
    assert get_url_from_entry(entry) == expected


def test_get_no_url_from_entry():
    bibtex = (
        '@ARTICLE{2009MNRAS.400.1181Z,\n'
        '  author = {{Zibetti}, S. and {Charlot}, S. and {Rix}, H.-W.},\n'
        '  title = "{Resolved stellar mass maps of galaxies - I. Method and '
        '  implications for global mass estimates}",\n'
        r'  journal = {\mnras},' + '\n'
        'archivePrefix = "arXiv",\n'
        '   eprint = {0904.4252},\n'
        ' primaryClass = "astro-ph.CO",\n'
        ' keywords = {techniques: image processing , '
        'techniques: photometric , galaxies: fundamental parameters , '
        'galaxies: general , galaxies: photometry , '
        'galaxies: stellar content},\n'
        '     year = 2009,\n'
        '    month = dec,\n'
        '   volume = 400,\n'
        '    pages = {1181-1198},\n'
        '  adsnote = {Provided by the SAO/NASA Astrophysics Data System}\n'
        '}\n'
    )
    db = parse_string(bibtex, 'bibtex')
    entry = db.entries['2009MNRAS.400.1181Z']
    with pytest.raises(NoEntryUrlError):
        get_url_from_entry(entry)


def test_get_authoryear():
    bibtex = (
        '@DocuShare{LDM-151,\n'
        '    author =       {John D. Swinbank and others},\n'
        '    title =        "{Data Management Science Pipelines Design}",\n'
        '    year =         2017,\n'
        '    month =        may,\n'
        '    handle =       {LDM-151},\n'
        '}\n'
    )
    db = parse_string(bibtex, 'bibtex')
    entry = db.entries['LDM-151']
    expected = 'Swinbank and others 2017'
    assert get_authoryear_from_entry(entry) == expected

    expected = 'Swinbank and others (2017)'
    assert get_authoryear_from_entry(entry, paren=True) == expected


def test_get_authoryear_single_author():
    bibtex = (
        '@DocuShare{LDM-151,\n'
        '    author =       {John D. Swinbank},\n'
        '    title =        "{Data Management Science Pipelines Design}",\n'
        '    year =         2017,\n'
        '    month =        may,\n'
        '    handle =       {LDM-151},\n'
        '}\n'
    )
    db = parse_string(bibtex, 'bibtex')
    entry = db.entries['LDM-151']
    expected = 'Swinbank 2017'
    assert get_authoryear_from_entry(entry) == expected

    expected = 'Swinbank (2017)'
    assert get_authoryear_from_entry(entry, paren=True) == expected


def test_get_authoryear_exception():
    # No year in this entry
    bibtex = (
        '@DocuShare{LDM-151,\n'
        '    author =       {John D. Swinbank},\n'
        '    title =        "{Data Management Science Pipelines Design}",\n'
        '    month =        may,\n'
        '    handle =       {LDM-151},\n'
        '}\n'
    )
    db = parse_string(bibtex, 'bibtex')
    entry = db.entries['LDM-151']
    with pytest.raises(AuthorYearError):
        get_authoryear_from_entry(entry)



"""Command line interface for lsstprojectmeta-ingest-docs.
"""
import argparse
import asyncio
import logging
import pprint

import aiohttp
from motor.motor_asyncio import AsyncIOMotorClient

from ..ltd import get_ltd_product_urls, get_ltd_product
from ..lsstdocument.handles import DOCUMENT_HANDLE_PATTERN
from ..lsstdocument.lander import process_lander_page, NotLanderPageError
from ..lsstdocument.sphinxtechnotes import (process_sphinx_technote,
                                            NotSphinxTechnoteError)


def main():
    """Command line entrypoint to reduce technote metadata.
    """
    parser = argparse.ArgumentParser(
        description='Discover and ingest metadata from document sources, '
                    'including lsstdoc-based LaTeX documents and '
                    'reStructuredText-based technotes. Metadata can be '
                    'upserted into the LSST Projectmeta MongoDB.')
    parser.add_argument(
        '--ltd-product',
        dest='ltd_product_url',
        help='URL of an LSST the Docs product '
             '(https://keeper.lsst.codes/products/<slug>). If provided, '
             'only this document will be ingested.')
    parser.add_argument(
        '--github-token',
        help='GitHub personal access token.')
    parser.add_argument(
        '--mongodb-uri',
        help='MongoDB connection URI. If provided, metadata will be loaded '
             'into the Projectmeta database. Omit this argument to just '
             'test the ingest pipeline.')
    parser.add_argument(
        '--mongodb-db',
        default='lsstprojectmeta',
        help='Name of MongoDB database')
    parser.add_argument(
        '--mongodb-collection',
        default='resources',
        help='Name of the MongoDB collection for projectmeta resources')
    args = parser.parse_args()

    # Configure the root logger
    stream_handler = logging.StreamHandler()
    stream_formatter = logging.Formatter(
        '%(asctime)s %(levelname)8s %(name)s | %(message)s')
    stream_handler.setFormatter(stream_formatter)
    root_logger = logging.getLogger()
    root_logger.addHandler(stream_handler)
    root_logger.setLevel(logging.WARNING)
    # Configure app logger
    app_logger = logging.getLogger('lsstprojectmeta')
    app_logger.setLevel(logging.DEBUG)

    if args.mongodb_uri is not None:
        mongo_client = AsyncIOMotorClient(args.mongodb_uri, ssl=True)
        collection = mongo_client[args.mongodb_db][args.mongodb_collection]
    else:
        collection = None

    loop = asyncio.get_event_loop()

    if args.ltd_product_url is not None:
        # Run single technote
        loop.run_until_complete(run_single_ltd_doc(args.ltd_product_url,
                                                   args.github_token,
                                                   collection))
    else:
        # Run bulk technote processing
        loop.run_until_complete(run_bulk_etl(args.github_token,
                                             collection))


async def run_single_ltd_doc(ltd_product_url, github_api_token,
                             mongo_collection):
    async with aiohttp.ClientSession() as session:
        jsonld = await process_ltd_doc(session, github_api_token,
                                       ltd_product_url,
                                       mongo_collection=mongo_collection)
    pp = pprint.PrettyPrinter(indent=2)
    pp.pprint(jsonld)


async def run_bulk_etl(github_api_token, mongo_collection):
    async with aiohttp.ClientSession() as session:
        product_urls = await get_ltd_product_urls(session)
        await process_ltd_doc_products(session, product_urls,
                                       github_api_token,
                                       mongo_collection=mongo_collection)


async def process_ltd_doc_products(session, product_urls, github_api_token,
                                   mongo_collection=None):
    """Run a pipeline to process extract, transform, and load metadata for
    multiple LSST the Docs-hosted projects

    Parameters
    ----------
    session : `aiohttp.ClientSession`
        Your application's aiohttp client session.
        See http://aiohttp.readthedocs.io/en/stable/client.html.
    product_urls : `list` of `str`
        List of LSST the Docs product URLs.
    github_api_token : `str`
        A GitHub personal API token. See the `GitHub personal access token
        guide`_.
    mongo_collection : `motor.motor_asyncio.AsyncIOMotorCollection`, optional
        MongoDB collection. This should be the common MongoDB collection for
        LSST projectmeta JSON-LD records.
    """
    tasks = [asyncio.ensure_future(
             process_ltd_doc(session, github_api_token,
                             product_url,
                             mongo_collection=mongo_collection))
             for product_url in product_urls]
    await asyncio.gather(*tasks)


async def process_ltd_doc(session, github_api_token, ltd_product_url,
                          mongo_collection=None):
    """Ingest any kind of LSST document hosted on LSST the Docs from its
    source.

    Parameters
    ----------
    session : `aiohttp.ClientSession`
        Your application's aiohttp client session.
        See http://aiohttp.readthedocs.io/en/stable/client.html.
    github_api_token : `str`
        A GitHub personal API token. See the `GitHub personal access token
        guide`_.
    ltd_product_url : `str`
        URL of the technote's product resource in the LTD Keeper API.
    mongo_collection : `motor.motor_asyncio.AsyncIOMotorCollection`, optional
        MongoDB collection. This should be the common MongoDB collection for
        LSST projectmeta JSON-LD records. If provided, ths JSON-LD is upserted
        into the MongoDB collection.

    Returns
    -------
    metadata : `dict`
        JSON-LD-formatted dictionary.

    .. `GitHub personal access token guide`: https://ls.st/41d
    """
    logger = logging.getLogger(__name__)

    ltd_product_data = await get_ltd_product(session, url=ltd_product_url)

    # Ensure the LTD product is a document
    product_name = ltd_product_data['slug']
    doc_handle_match = DOCUMENT_HANDLE_PATTERN.match(product_name)
    if doc_handle_match is None:
        logger.debug('%s is not a document repo', product_name)
        return

    # Figure out the format of the document by probing for metadata files.
    # reStructuredText-based Sphinx documents have metadata.yaml file.
    try:
        return await process_sphinx_technote(session,
                                             github_api_token,
                                             ltd_product_data,
                                             mongo_collection=mongo_collection)
    except NotSphinxTechnoteError:
        # Catch error so we can try the next format
        logger.debug('%s is not a Sphinx-based technote.', product_name)
    except Exception:
        # Something bad happened trying to process the technote.
        # Log and just move on.
        logger.exception('Unexpected error trying to process %s', product_name)
        return

    # Try interpreting it as a Lander page with a /metadata.jsonld document
    try:
        return await process_lander_page(session,
                                         github_api_token,
                                         ltd_product_data,
                                         mongo_collection=mongo_collection)
    except NotLanderPageError:
        # Catch error so we can try the next format
        logger.debug('%s is not a Lander page with a metadata.jsonld file.',
                     product_name)
    except Exception:
        # Something bad happened; log and move on
        logger.exception('Unexpected error trying to process %s', product_name)
        return



"""LaTeX macros provided by the lsstdoc class.

lsstprojectmeta mirrors these macros to allow Pandoc to resolve them.
"""

__all__ = ['LSSTDOC_MACROS']

LSSTDOC_MACROS = r"""\newcommand{\latex}{LaTeX}
\newcommand{\docType}{LSST Document}
\newcommand{\CU}{DM}
\newcommand{\DU}{}
\newcommand{\setLSSTDU}[1]{\renewcommand{\DU}{-#1}}
\newcommand{\byte}{b}
\def\aaps{A\&AS}           % Astronomy and Astrophysics Suplement
\def\aap{A\&A}             % Astronomy and Astrophysics
\def\ssr{Space~Sci.~Rev.}  % Space Science Reviews
\def\apj{ApJ}              % Astrophysical Journal
\def\apjs{ApJS}            % Astrophysical Journal Supplement
\def\aj{AJ}                % Astronomical Journal
\def\mnras{MNRAS}          % Monthly Notices of the RAS
\def\araa{ARA\&A}          % Annual Review of Astron and Astrophys
\def\nat{Nature}           % Nature
\def\apjl{ApJ}             % Astrophysical Journal, Letters
\def\icarus{Icarus}        % Icarus
\def\prd{Phys.~Rev.~D}     % Physical Review D
\def\physrep{Phys.~Rep.}   % Physics Reports
\def\pasp{PASP}            % Publications of the Astronomical Society of the Pacific
\def\procspie{Proc.\ SPIE} % Proceedings of the SPIE
\newcommand{\pasa}{PASA}   % Publications of the Astronomical Society of Australia
\newcommand{\ao}{Appl.~Opt.}  % Applied Optics

\def\degr{\hbox{$^\circ$}}
\def\arcmin{\hbox{$^\prime$}}
\def\arcsec{\hbox{$^{\prime\prime}$}}
\def\fs{\hbox{$.\!\!^{\rm s}$}}
\def\fdg{\hbox{$.\!\!^\circ$}}
\def\farcm{\hbox{$.\mkern-4mu^\prime$}}
\def\farcs{\hbox{$.\!\!^{\prime\prime}$}}
\def\sun{\hbox{$\odot$}}
\newcommand{\pu}[2]{\ensuremath{#1\,\mbox{#2}}}
\newcommand{\secs}[1]{\pu{#1}{s}}
\newcommand{\uas}[1]{\pu{#1}{$\mu$as}}
\newcommand{\muas}[1]{\pu{#1}{$\mu$as}}
\newcommand{\muasonly}{\ensuremath{\mu\mbox{as}}}
\newcommand{\inp}[1]{{\tt #1}}
\newcommand{\prog}[1]{{\it #1}}
\newcommand{\secref}[1]{Section~\ref{#1}}
\newcommand{\appref}[1]{Appendix~\ref{#1}}
\newcommand{\partref}[1]{Part~\ref{#1}}
\newcommand{\tabref}[1]{Table~\ref{#1}}
\newcommand{\figref}[1]{Figure~\ref{#1}}
\newcommand{\eqnref}[1]{Eq.~\ref{#1}}
\newcommand{\reqref}[1]{Req.~\ref{#1}}
\newcommand{\actref}[1]{AI~\ref{#1}}
\newcommand{\class}[1]{{\it #1}}
\newcommand{\pack}[1]{{\em #1}}
\newcommand{\property}[1]{{\em #1}}
\newcommand{\method}[1]{{\em #1}}
\newcommand{\file}[1]{{\em #1}}
\newcommand{\directory}[1]{{\em #1}}
%\providecommand{\code}[1]{{\tt #1}}

% \renewcommand{\vec}[1]{\ensuremath{\mathchoice{\mbox{\boldmath$\displaystyle#1$}}
% {\mbox{\boldmath$\textstyle#1$}}
% {\mbox{\boldmath$\scriptstyle#1$}}
% {\mbox{\boldmath$\scriptscriptstyle#1$}}}}

\newcommand{\vstretch}[1]{\vspace*{\stretch{#1}}}
\newcommand{\code}[1]{\texttt{#1}}

\newcommand{\Alert}{\code{Alert}\xspace}
\newcommand{\Alerts}{\code{Alerts}\xspace}
\newcommand{\DIASource}{\code{DIA\-Source}\xspace}
\newcommand{\DIASources}{\code{DIA\-Sources}\xspace}
\newcommand{\DIAObject}{\code{DIA\-Object}\xspace}
\newcommand{\DIAObjects}{\code{DIA\-Objects}\xspace}
\newcommand{\DB}{{Level 1 database}\xspace}
\newcommand{\DR}{{Level 2 database}\xspace}
\newcommand{\Object}{\code{Object}\xspace}
\newcommand{\Objects}{\code{Objects}\xspace}
\newcommand{\Source}{\code{Source}\xspace}
\newcommand{\Sources}{\code{Sources}\xspace}
\newcommand{\ForcedSource}{\code{Forced\-Source}\xspace}
\newcommand{\ForcedSources}{\code{Forced\-Sources}\xspace}
\newcommand{\CoaddSource}{\code{Coadd\-Source}\xspace}
\newcommand{\CoaddSources}{\code{Coadd\-Sources}\xspace}
\newcommand{\SSObject}{\code{SS\-Object}\xspace}
\newcommand{\SSObjects}{\code{SS\-Objects}\xspace}
\newcommand{\VOEvent}{\code{VO\-Event}\xspace}
\newcommand{\VOEvents}{\code{VO\-Events}\xspace}
\newcommand{\transSNR}{5\xspace}

% Command to link to a document in Docushare. Pass an LSST document handle as argument, or a document number.
% This will not result in a references entry.
\newcommand{\ds}[2]{{\color{blue} \href{https://docushare.lsst.org/docushare/dsweb/Get/#1}{#2}}\xspace}

% These used to be `citeds` commands. Now just plain text with hyperlinks.
\newcommand{\SRD}{\href{https://docushare.lsst.org/docushare/dsweb/Get/LPM-17}{SRD}\xspace}
\newcommand{\DPDD}{\href{https://docushare.lsst.org/docushare/dsweb/Get/LSE-163}{DPDD}\xspace}
\newcommand{\LSR}{\href{https://docushare.lsst.org/docushare/dsweb/Get/LSE-29}{LSR}\xspace}
\newcommand{\OSS}{\href{https://docushare.lsst.org/docushare/dsweb/Get/LSE-30}{OSS}\xspace}
\newcommand{\DMSR}{\href{https://docushare.lsst.org/docushare/dsweb/Get/LSE-61}{DMSR}\xspace}
\newcommand{\appsUMLdomain}{https://docushare.lsst.org/docushare/dsweb/Get/\href{LDM-133}{LDM-133}\xspace}
\newcommand{\appsUMLusecase}{\href{https://docushare.lsst.org/docushare/dsweb/Get/LDM-134}{LDM-134}\xspace}
\newcommand{\SUI}{\href{https://docushare.lsst.org/docushare/dsweb/Get/LDM-131}{SUID}\xspace}
\newcommand{\DMSD}{\href{https://docushare.lsst.org/docushare/dsweb/Get/LDM-148}{DMSD}\xspace}
\newcommand{\MOPSD}{\href{https://docushare.lsst.org/docushare/dsweb/Get/LDM-156}{MOPSD}\xspace}
\newcommand{\DMMD}{\href{https://docushare.lsst.org/docushare/dsweb/Get/LDM-152}{DMMD}\xspace}
\newcommand{\DMOps}{\href{https://docushare.lsst.org/docushare/dsweb/Get/LDM-230}{DM OpsCon}\xspace}
\newcommand{\SDQAP}{\href{https://docushare.lsst.org/docushare/dsweb/Get/LSE-63}{LSE-63}\xspace}
\newcommand{\NewPCP}{\href{https://docushare.lsst.org/docushare/dsweb/Get/LSE-180}{LSE-180}\xspace}
\newcommand{\UCAL}{\href{https://docushare.lsst.org/docushare/dsweb/Get/Document-15125}{UCAL}\xspace}

\newcommand{\wbsSFM}{WBS 02C.03.01}
\newcommand{\wbsAssocP}{WBS 02C.03.02}
\newcommand{\wbsAP}{WBS 02C.03.03}
\newcommand{\wbsDiffim}{WBS 02C.03.04}
\newcommand{\wbsMOPS}{WBS 02C.03.06}
\newcommand{\wbsSDQAP}{WBS 02C.01.02.02}
\newcommand{\wbsSDQAT}{WBS 02C.01.02.02}
\newcommand{\wbsSPT}{WBS 02C.01.02.03}
\newcommand{\wbsPSF}{WBS 02C.04.03}
\newcommand{\wbsCoadd}{WBS 02C.04.04}
\newcommand{\wbsDetDeblend}{WBS 02C.04.05}
\newcommand{\wbsObjChar}{WBS 02C.04.06}
\newcommand{\wbsAFW}{WBS 02C.03.05, 02C.04.01}
\newcommand{\wbsCPP}{WBS 02C.04.02}
\newcommand{\wbsPhotoCal}{WBS 02C.03.07}
\newcommand{\wbsAstroCal}{WBS 02C.03.08}

%%% Command's for merging Robert's document, at least temporarily.
%%% Some of these will be removable, some may not be.
\renewcommand{\c}{\textit{c.}\xspace}
\newcommand{\eg}{\textit{e.g.}\xspace}
\newcommand{\etc}{\textit{etc.}\xspace}
\newcommand{\ie}{\textit{i.e.}\xspace}
\renewcommand{\th}{\ensuremath{^{\mbox{\tiny th}}}\xspace}
\newcommand{\Nb}{\textit{N.b.}\xspace}
\newcommand{\nb}{\textit{n.b.}\xspace}
\DeclareBoldMathCommand\xib{\xi}
\DeclareBoldMathCommand\zerob{0}
\newcommand{\cameraTeam}{\textbf{Camera Team deliverable}\\}
\newcommand{\TBD}{\textit{T.B.D.}\xspace}
\newcommand{\myLabel}[2]{\def\@currentlabel{#2}\label{#1}}
\newcommand{\appRef}[1]{App. \ref{app:#1}}
\newcommand{\secRef}[1]{Sec. \ref{sec:#1}}
\newcommand{\rednote}[1]{{\color{red} (#1)}}

\newcommand{\alg}{\\ \textbullet\xspace Algorithmic component: }
\newcommand{\fefiftyfive}{$^{55}$Fe\xspace}
\newcommand{\bfeffect}{``brighter-fatter effect"}
\newcommand{\auxtelescope}{auxiliary/Calpyso/calibration telescope\xspace}
\newcommand{\secsymbol}{§}
\newcommand{\xxx}{{\color{red} xxx}\xspace}
\newcommand{\dragons}{ {\color{blue} \textbf{\textit{Here be dragons...}}}}
\newcommand{\mysubitem}{\subitem \textendash\xspace}
\newcommand{\smalltilde}{\raise.26ex\hbox{$\scriptstyle\mathtt{\sim}$}}
\newcommand{\electron}{$e^{-}$}

\newcommand{\oldtext}[1]{{\color{red} #1}}
\newcommand{\newtext}[1]{{\color{blue} #1}}

\newcommand{\jira}[1]{\href{https://jira.lsstcorp.org/browse/#1}{#1}}

\newcommand{\uc}[1]{\texttt{#1}}
"""



"""Test the lsstprojectmeta-deparagraph Pandoc filter.

The filter is implemented in lsstprojectmeta.pandoc.filters.deparagraph, but we
test it through the lsstprojectmeta-deparagraph entrypoint.
"""

import pypandoc
import pytest


@pytest.mark.parametrize(
    'sample,expected', [
        # Should strip <p> tag from single paragraph.
        ('Hello world!', 'Hello world!\n'),
        # Should leave <p> tags for multiple paragraphs.
        ('Hello.\n\nWorld!', '<p>Hello.</p>\n<p>World!</p>\n')
    ])
def test_deparagraph(sample, expected):
    output = pypandoc.convert_text(
        sample, 'html5', format='latex',
        extra_args=['--filter=lsstprojectmeta-deparagraph'])
    assert output == expected



from setuptools import setup, find_packages
import os
from io import open


packagename = 'lsst-projectmeta-kit'
description = ("Python toolkit for extracting and transforming metadata "
               "about LSST's code and documentation projects, and loading "
               "it into the LSST projectmeta database.")
author = 'Association of Universities for Research in Astronomy, Inc.'
author_email = 'jsick@lsst.org'
license = 'MIT'
url = 'https://github.com/lsst-sqre/lsst-projectmeta-kit'


def read(filename):
    full_filename = os.path.join(
        os.path.abspath(os.path.dirname(__file__)),
        filename)
    return open(full_filename, mode='r', encoding='utf-8').read()


long_description = read('README.rst')


setup(
    name=packagename,
    use_scm_version=True,
    description=description,
    long_description=long_description,
    url=url,
    author=author,
    author_email=author_email,
    license=license,
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='lsst',
    packages=find_packages(exclude=['docs', 'tests*', 'data']),
    install_requires=[
        'pyjwt>=1.4.2',
        'requests>=2.13.0',
        'pypandoc>=1.4',
        'panflute==1.10.6',
        'aiohttp>=2.2.5',
        'pybtex>=0.21',
        'GitPython>=2.1.7',
        'pytz',
        'motor>=1.2.0, <1.3.0',
        # https://github.com/Julian/jsonschema/issues/449#issuecomment-411406525
        'attrs>=17.4.0 '
    ],
    extras_require={
        'dev': [
            # Development/testing dependencies
            'pytest==4.4.0',
            'pytest-cov==2.6.1',
            'pytest-flake8==1.0.4',
        ]},
    setup_requires=[
        'setuptools-scm==1.15.6',
    ],
    include_package_data=True,
    entry_points={
        'console_scripts': [
            ('lsstprojectmeta-deparagraph '
             '= lsstprojectmeta.pandoc.filters.deparagraph:main'),
            ('lsstprojectmeta-ingest-docs '
             '= lsstprojectmeta.cli.ingestdocs:main')
        ]
    }
)



"""Extract metadata from document landing pages created by Lander_.

.. Lander: https://github.com/lsst/lander
"""

__all__ = ('process_lander_page', 'NotLanderPageError')

import logging
from urllib.parse import urljoin

import aiohttp

from ..jsonld import decode_jsonld


async def process_lander_page(session, github_api_token, ltd_product_data,
                              mongo_collection=None):
    """Extract, transform, and load metadata from Lander-based projects.

    Parameters
    ----------
    session : `aiohttp.ClientSession`
        Your application's aiohttp client session.
        See http://aiohttp.readthedocs.io/en/stable/client.html.
    github_api_token : `str`
        A GitHub personal API token. See the `GitHub personal access token
        guide`_.
    ltd_product_data : `dict`
        Contents of ``metadata.yaml``, obtained via `download_metadata_yaml`.
        Data for this technote from the LTD Keeper API
        (``GET /products/<slug>``). Usually obtained via
        `lsstprojectmeta.ltd.get_ltd_product`.
    mongo_collection : `motor.motor_asyncio.AsyncIOMotorCollection`, optional
        MongoDB collection. This should be the common MongoDB collection for
        LSST projectmeta JSON-LD records. If provided, ths JSON-LD is upserted
        into the MongoDB collection.

    Returns
    -------
    metadata : `dict`
        JSON-LD-formatted dictionary.

    Raises
    ------
    NotLanderPageError
        Raised when the LTD product cannot be interpreted as a Lander page
        because the ``/metadata.jsonld`` file is absent. This implies that
        the LTD product *could* be of a different format.

    .. `GitHub personal access token guide`: https://ls.st/41d
    """
    logger = logging.getLogger(__name__)

    # Try to download metadata.jsonld from the Landing page site.
    published_url = ltd_product_data['published_url']
    jsonld_url = urljoin(published_url, '/metadata.jsonld')
    try:
        async with session.get(jsonld_url) as response:
            logger.debug('%s response status %r', jsonld_url, response.status)
            response.raise_for_status()
            json_data = await response.text()
    except aiohttp.ClientResponseError as err:
        logger.debug('Tried to download %s, got status %d',
                     jsonld_url, err.code)
        raise NotLanderPageError()
    # Use our own json parser to get datetimes
    metadata = decode_jsonld(json_data)

    if mongo_collection is not None:
        await _upload_to_mongodb(mongo_collection, metadata)

    return metadata


async def _upload_to_mongodb(collection, jsonld):
    """Upsert the technote resource into the projectmeta MongoDB collection.

    Parameters
    ----------
    collection : `motor.motor_asyncio.AsyncIOMotorCollection`
        The MongoDB collection.
    jsonld : `dict`
        The JSON-LD document that represents the document resource.
    """
    document = {
        'data': jsonld
    }
    query = {
        'data.reportNumber': jsonld['reportNumber']
    }
    await collection.update(query, document, upsert=True, multi=False)


class NotLanderPageError(Exception):
    """Exception indicating that an LSST the Docs product cannot be a Lander
    landing page because ``/metadata.jsonld`` is missing.
    """



"""Pandoc filter to remove the outer Para wrapper and replace it with a
Plain wrapper.
"""

from panflute import toJSONFilter, Para, Plain


def deparagraph(element, doc):
    """Panflute filter function that converts content wrapped in a Para to
    Plain.

    Use this filter with pandoc as::

        pandoc [..] --filter=lsstprojectmeta-deparagraph

    Only lone paragraphs are affected. Para elements with siblings (like a
    second Para) are left unaffected.

    This filter is useful for processing strings like titles or author names so
    that the output isn't wrapped in paragraph tags. For example, without
    this filter, pandoc converts a string ``"The title"`` to
    ``<p>The title</p>`` in HTML. These ``<p>`` tags aren't useful if you
    intend to put the title text in ``<h1>`` tags using your own templating
    system.
    """
    if isinstance(element, Para):
        # Check if siblings exist; don't process the paragraph in that case.
        if element.next is not None:
            return element
        elif element.prev is not None:
            return element

        # Remove the Para wrapper from the lone paragraph.
        # `Plain` is a container that isn't rendered as a paragraph.
        return Plain(*element.content)


def main():
    """Setuptools entrypoint for the deparagraph CLI.

    Use this filter as::

        pandoc [..] --filter=lsstprojectmeta-deparagraph
    """
    toJSONFilter(deparagraph)



"""JSON-LD utilities.
"""

__all__ = ('encode_jsonld', 'JsonLdEncoder', 'decode_jsonld')

import datetime
import json


def encode_jsonld(jsonld_dataset, **kwargs):
    """Encode a JSON-LD dataset into a string.

    Parameters
    ----------
    jsonld_dataset : `dict`
        A JSON-LD dataset.
    kwargs
        Keyword argument passed to the encoder. See `json.JSONEncoder`.

    Returns
    -------
    encoded_dataset : `str`
        The JSON-LD dataset encoded as a string.
    """
    encoder = JsonLdEncoder(**kwargs)
    return encoder.encode(jsonld_dataset)


class JsonLdEncoder(json.JSONEncoder):
    """Customized JSON encoder (replaces `json.JSONEncoder`) that supports
    datetime encoding.
    """

    def default(self, obj):
        """Encode values as JSON strings.

        This method overrides the default implementation from
        `json.JSONEncoder`.
        """
        if isinstance(obj, datetime.datetime):
            return self._encode_datetime(obj)

        # Fallback to the default encoding
        return json.JSONEncoder.default(self, obj)

    def _encode_datetime(self, dt):
        """Encode a datetime in the format '%Y-%m-%dT%H:%M:%SZ'.

        The datetime can be naieve (doesn't have timezone info) or aware
        (it does have a tzinfo attribute set). Regardless, the datetime
        is transformed into UTC.
        """
        if dt.tzinfo is None:
            # Force it to be a UTC datetime
            dt = dt.replace(tzinfo=datetime.timezone.utc)

        # Convert to UTC (no matter what)
        dt = dt.astimezone(datetime.timezone.utc)

        return dt.strftime('%Y-%m-%dT%H:%M:%SZ')


def decode_jsonld(jsonld_text):
    """Decode a JSON-LD dataset, including decoding datetime
    strings into `datetime.datetime` objects.

    Parameters
    ----------
    encoded_dataset : `str`
        The JSON-LD dataset encoded as a string.

    Returns
    -------
    jsonld_dataset : `dict`
        A JSON-LD dataset.

    Examples
    --------

    >>> doc = '{"dt": "2018-01-01T12:00:00Z"}'
    >>> decode_jsonld(doc)
    {'dt': datetime.datetime(2018, 1, 1, 12, 0, tzinfo=datetime.timezone.utc)}
    """
    decoder = json.JSONDecoder(object_pairs_hook=_decode_object_pairs)
    return decoder.decode(jsonld_text)


def _decode_object_pairs(pairs):
    doc = {}
    for key, value in pairs:
        if isinstance(value, str):
            # attempt to parse as a datetime
            try:
                value = datetime.datetime.strptime(value, '%Y-%m-%dT%H:%M:%SZ')
                # ensure timezone is UTC
                if value.tzinfo is None:
                    value = value.replace(tzinfo=datetime.timezone.utc)
                value = value.astimezone(datetime.timezone.utc)
            except ValueError:
                pass

        doc[key] = value
    return doc
